/*
 PPU.c
  Picture Processing Unit 2C0x-level code

 * MacFCEU 0.9a2
 *
 * FCE Ultra - NES/Famicom Emulator
 *
 * Copyright notice for this file:
 *  ppu.c Copyright (C) 1998 BERO
 *  ppu.c Copyright (C) 2003 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "types.h"
#include "x65xx.h"
#include "FCEU.h"
#include "git.h"
#include "PPU.h"
#include "NSF.h"
#include "GameGenie.h"
#include "fsound.h"
#include "general.h"
#include "memory.h"

#include "cart.h"
#include "palette.h"
#include "state.h"
#include "video.h"
#include "input.h"
#include "boards/MMC5.h"

#include "ftrig.h"

#ifdef env68k
 #pragma segment shell
#endif

//#undef env68k
#include <stdlib.h>
/* ---- constants */

/* The makeup of the video 15 bit daisy chain:
 bits 12 : fv.3	- fine vertical scroll
 bit  11 : v.1	- vertical scroll counter
 bit  10 : h.1	- horizontal scroll counter
 bits  5 : vt.5	- vertical tile position   -- bit #1 of this shifts palette by 4
 bits  0 : ht.5	- horizontal tile position -- bit #1 of this shifts palette by 2
 bits(-3): fh.3	- fine horizontal scroll
*/
#define HV_SCR_SHIFT		10
#define HT_POS_MASK		0x1F
#define H_POS_MASKS		0x41F
#define T_POS_MASKS		0x3FF
#define POS_MASKS		0x3FFF

#define STAT_OAM_OV		0x20
#define STAT_OAM_ZH		0x40
#define STAT_VCLR		0x77
#define STAT_VINT		0x80
#define STAT_RDMASK		0xE0

/* CPU subcycle time scales */
#define PAL_SC			(CPU_SUBCYC*5/16)
#define NTSC_SC			(CPU_SUBCYC*4/12)

/* number of cycles spent on the boot line before sleeping for 240 lines */
#define BOOTSYNCCYC		312

/* really 1, but we'd have to account for write-cycle timing, etc. */
#define NMI_DISABLE_DELAY	CPUSC[PPU_2]

#define CPU_NMI_SC		(CPU_SUBCYC*CPU_NMI_DELAY_CYC+CPU_INTEDGE)

/* ---- globals */

int		fceuindbg = 0;

uint8		PPUmaxsprites = 8;
uint8		PPUrendis;
uint8		PPUVFirstField, PPUBurstCount;

uint32		PPUVFrameCounter;
uint16		PPUChipType;
static uint8	A02Version, A02LatMask;
int8		PPUBoot, PPURegBoot;

uint8		VRAMBuffer, PPUGenLatch, PPULatDecay;
uint8		*vnapage[4];
uint8		PPUNTARAM;
uint8		PPUCHRRAM;

/* Color deemphasis emulation.  Joy... */
uint32		PPUEmphasis = 0;

void (*GameHBIRQHook)(void), (*GameHBIRQHook2)(void);
void FP_FASTAPASS(1) (*PPU_hook)(uint32 A);
void (*PPU_A12hook)(void);

uint32		TempAddr, RefreshAddr;

/* "scanline" is the current visible scanline we're on */

int		scanline;

uint8		PPUReg[4];
uint8		PPUSPL, PPUSPO, OAMLatch;
uint8		vtoggle;
uint32		XOffset;

uint8		PPUvoidcolor = 255;

uint16		EmphLUT[8]; /* region-specific precalculated 11-bits */

/* new form: a modified *copy* of PALRAM is used; no more hacks. :) */
static uint8	PALRAMMask;
PLB_pel		PPUPal[32];
uint8		PALRAM[32], NTARAM[0x800];

/* TODO: rewrite OAM code */
/* NTSC OAM RAM decays to zero when rendering is off; PAL does not */
uint8		SPRAM[0x104], SPRBUF[0x104];

static uint64	A2007nexttime;

static PLB_pel	PPU_Plinef[352];
PLB_pel		*PPU_Pline;
static int	linepel;
static int	linestartts;
static uint8	linereset;//, NextFirst8;

uint8		bitreverseLUT[256];
#ifndef env68k
uint32		PPU_LUT1to4[256]; /* 1024B */
uint32		PPU_ATLUT[128]; /* 512B */
#endif

/* prefetching */
static uint32	pshift0, pshift1;
static uint32	atlatch;

/* CPU subcycle timings */

enum
{
	PPU_VBL,
	PPU_1,
	PPU_2,
	PPU_3,
	PPU_4,
	PPU_5,
	PPU_6,
	PPU_12,
	PPU_16,
	PPU_54,
	PPU_58,
	PPU_64,
	PPU_85,
	PPU_192,
	PPU_256,
	PPU_SL,
	PPU_odd, /* NTSC odd frames drop one cycle */
	PPU_NMI_M2,
	PPU_EFSleep, /* end-frame sleep time */
	PPU_BootSync,
	PPU_visible,
	PPU_timings,

	PPU_even = PPU_4
};

int		CPUSC[PPU_timings];

/* ---- macros */

#define VBlankON		(PPUReg[0]&128)	/* Generate VBlank NMI */
#define Sprite16		(PPUReg[0]&32)	/* Sprites 8x16/8x8 */
#define BGAdrHI			(PPUReg[0]&16)	/* PF pattern adr $0000/$1000 */
#define SpAdrHI			(PPUReg[0]&8)	/* OAM pattern adr $0000/$1000 */
#define IncByV			(PPUReg[0]&4)	/* by 32, otherwise 1 (horiz.) */

#define RenderON		(PPUReg[1]&24)	/* either */
#define SpriteON		(PPUReg[1]&16)	/* show OAM layer */
#define PlayfdON		(PPUReg[1]&8)	/* show playfield */
#define SpFirst8		(PPUReg[1]&4)	/* render first 8 OAM pixels on line */
#define BGFirst8		(PPUReg[1]&2)	/* render first 8 PF pixels on line */
#define GrayLine		(PPUReg[1]&1)	/* erase line color */

#define PPU_status		PPUReg[2]

#define PPUNoFrame		(scanline>240||!RenderON)

#define exp8(c)			(((PLB_pel)(c)*0x101u)*0x10001u)
#ifdef C80x86
 #define tilespr32(P,R,d)	do{PLB_pel *S=R; uint32 V=d; \
	 asm volatile( \
	 "movl %%ebx,%%eax\n\t" \
	 "andl $15,%%eax\n\t" \
	 "movb (%%esi,%%eax),%%cl\n\t" \
	 "shrl $4, %%ebx\n\t" \
	 "movb %%cl, (%%edi)\n\t" \
\
         "movl %%ebx,%%eax\n\t" \
         "andl $15,%%eax\n\t" \
         "movb (%%esi,%%eax),%%cl\n\t" \
         "shrl $4, %%ebx\n\t" \
         "movb %%cl, 1(%%edi)\n\t" \
\
         "movl %%ebx,%%eax\n\t" \
         "andl $15,%%eax\n\t" \
         "movb (%%esi,%%eax),%%cl\n\t" \
         "shrl $4, %%ebx\n\t" \
         "movb %%cl, 2(%%edi)\n\t" \
\
         "movl %%ebx,%%eax\n\t" \
         "andl $15,%%eax\n\t" \
         "movb (%%esi,%%eax),%%cl\n\t" \
         "shrl $4, %%ebx\n\t" \
         "movb %%cl, 3(%%edi)\n\t" \
\
         "movl %%ebx,%%eax\n\t" \
         "andl $15,%%eax\n\t" \
         "movb (%%esi,%%eax),%%cl\n\t" \
         "shrl $4, %%ebx\n\t" \
         "movb %%cl, 4(%%edi)\n\t" \
\
         "movl %%ebx,%%eax\n\t" \
         "andl $15,%%eax\n\t" \
         "movb (%%esi,%%eax),%%cl\n\t" \
         "shrl $4, %%ebx\n\t" \
         "movb %%cl, 5(%%edi)\n\t" \
\
         "movl %%ebx,%%eax\n\t" \
         "andl $15,%%eax\n\t" \
         "movb (%%esi,%%eax),%%cl\n\t" \
         "shrl $4, %%ebx\n\t" \
         "movb %%cl, 6(%%edi)\n\t" \
\
         /*"movl %%ebx,%%eax\n\t"*/ \
         /*"andl $15,%%eax\n\t"*/ \
         /*"movb (%%esi,%%eax),%%cl\n\t"*/ \
         /*"movb %%cl, 7(%%edi)\n\t"*/ \
	 "movb (%%esi, %%ebx),%%cl\n\t" \
	 "movb %%cl, 7(%%edi)\n\t" \
	 : \
	 :"S" (S), "D" (P), "b" (V) \
	 :"eax", "cl"); \
}while(0)
#elif defined(env68k)
 #define blittile(d,p,m,a,x)	_blittile(d,PPUPal,(unsigned)p<<16|(uint16)m,a,x)
 #pragma parameter __A0 _blittile(__A0, __A1, __D2, __D1, __D0)
 extern PLB_pel *_blittile(PLB_pel *D,
	const void *T,
	uint32 P,
	uint16 A,
	uint16 X) = {
/*	MOVE.W		D1,-(SP) */ 0x3F01,
/*	MOVE.W		D0,-(SP) */ 0x3F00,
/*	LSL.L		D0,D2	*/ 0xE1AA,
/*	MOVE.L		D2,D1	*/ 0x2202,
/*	CLR.W		D1	*/ 0x4241,
/*	SUBQ.B		#8,D0	*/ 0x5100,
/*@0	MOVE.B		3(A7),D1 */ 0x122F, 3,
/*	ADD.W		D2,D2	*/ 0xD442,
/*	ADDX.L		D1,D1	*/ 0xD381,
/*	ADDX.B		D1,D1	*/ 0xD301,
/*	MOVE.B		0(A1,D1.W),(A0)+ */ 0x10F1, 0x1000,
/*	ADDQ.B		#1,D0	*/ 0x5200,
/*	BCC.S		@0	*/ 0x64EE,
/*	MOVE.W		(A7)+,D0 */ 0x301F,
/*	BEQ.S		@2	*/ 0x6710,
/*@1	MOVE.B		(A7),D1	*/ 0x1217,
/*	ADD.W		D2,D2	*/ 0xD442,
/*	ADDX.L		D1,D1	*/ 0xD381,
/*	ADDX.B		D1,D1	*/ 0xD301,
/*	MOVE.B		0(A1,D1.W),(A0)+ */ 0x10F1, 0x1000,
/*	SUBQ.B		#1,D0	*/ 0x5300,
/*	BNE.S		@1	*/ 0x66F0,
/*@2	ADDQ.W		#2,A7	*/ 0x544F};
#else
 #define tilespr32(P,R,d)	do{ \
	PLB_pel *S=R; uint32 V=d; \
	*P=S[V&15]; \
	P[1]=S[(V>>=4)&15]; \
	P[2]=S[(V>>=4)&15]; \
	P[3]=S[(V>>=4)&15]; \
	P[4]=S[(V>>=4)&15]; \
	P[5]=S[(V>>=4)&15]; \
	P[6]=S[(V>>=4)&15]; \
	P[7]=S[V>>=4]; \
}while(0)
#endif

#define MMC5SPRVRAMADR(A)	&MMC5SPRVPage[(A)>>HV_SCR_SHIFT][A]
#define MMC5BGVRAMADR(A)	(!Sprite16 && !MMC5ABMode ? \
	&MMC5SPRVPage[A>>10][A]:&MMC5BGVPage[A>>10][A])
#define VRAMADR(A)		&VPage[(A)>>HV_SCR_SHIFT][A]

#define VAddr(a) \
	((a)<0x2000 ? VPage[(a)>>HV_SCR_SHIFT]+a:vnapage[(a)>>HV_SCR_SHIFT&3]+((a)&0x3FF))
static uint8 *VAddrSafe(a)
{
	uint8		*p;

	if (a<0x2000)
		p = VPage[(a)>>HV_SCR_SHIFT];
	else if (a<0x3F00 || FCEUGameInfo->vidsys==GIV_RGB)
		p = vnapage[(a)>>HV_SCR_SHIFT&3], a&= 0x3FF;
	else if (a<0x4000)
	{
		if (!(a&=31))
			a = 16;
		p = PALRAM;
	} else
		return 0;
	return p ? p+a:0;
}

/*
 ppu_open_bus test shows the I/O latch decays to 0 in about 0.6 sec.
  Perform at least one check per frame to clear these bits past a minimum time.
  Remember that the top two bits for the palette RAM are from the "decay" bits.
  Also, patterns like AAh and 55h may cause the port latch to decay faster,
  but that'd seem a bit excessive to implement here.  Thanks to wiki.nesdev.com.
*/
/* 33 = somewhere between 1/2 and 2/3 sec for both NTSC and PAL */
#define SetDecayLatch(V) (void)(PPUGenLatch=V,PPULatDecay=33)

/* ---- code */

static void makeluts(
	uint8 *brt,
#ifndef env68k
	uint32 *l1b,
	uint32 *tlu
#endif
	)
{
	{
		uint8	*d = brt, v;
		uint32	x = 0;

		do
		{
			v = x<<5&0x40;
			v|= x<<3&0x20;
			v|= (x+x)&0x10;
			v|= x>>1&8;
			v|= x>>3&4;
			v|= x>>5&2;
			v|= x>>7;
			*d++ = v, v|= 0x80;
			*d++ = v;
		} while((x+=2)<256);
	}
#ifndef env68k
	{
		uint32	*d = l1b, x = 0, v;

		do
		{ /* unrolled */
			v = x>>7;
			v|= x>>6-4&0x10;
			v|= x<<8-5&0x100;
			v|= x<<12-4&0x1000;
			v|= x<<16-3&0x10000;
			v|= x<<20-2&0x100000;
			v|= x<<24-1&0x1000000;
			*d++ = v,v|= 0x10000000, *d++ = v;
		} while((x+=2)<256);
	}
	{
		uint32	*d = tlu, cc = 0, xo;

		do
		{
			xo = 0;
			do
			{
				uint32	i = 8, k = xo, v = 0, a = 2;

				do
					v|= ((cc>>(k>>3)*2)&3)<<a;
				while(++k,a+=4,--i);
				*d++=v;
			} while(++xo<8);
		} while(++cc<16);
	}
#endif
}

DECLFW(PPUB2004);

#ifdef env68k
DECLFR(PPUA2002);
DECLFR(PPUA200x);
DECLFR(PPUA2007);

DECLFW(PPUB2000);
DECLFW(PPUB2001);
DECLFW(PPUB2002);
DECLFW(PPUB2003);
DECLFW(PPUB2005);
DECLFW(PPUB2006);
DECLFW(PPUB2007);

void CheckSpriteHit(int e);
void RefreshSprites(void);
 #define STATIC
#else
 #define STATIC static
#endif

static void FASTAPASS(1) RefreshLine(int lastpixel);

#define LinePixel()	(PAL ? /* /15 multiply optimization; 2011.02.08 */ \
	ufract32mul(timestamp*CPU_SUBCYC-linestartts,(4294967296.*8-1+PAL_SC)/PAL_SC)>>3 \
	:(timestamp*CPU_SUBCYC-linestartts)/NTSC_SC)

int FCEUPPU_GetPixel(void)
{
	if (PPU_Pline)
		return LinePixel()%LCYCS;
	return -1;
}

uint8 FCEUPPU_MemPeek(uint16 a)
{
	const uint8	*p;

	if ((p=VAddrSafe(a&=0x3FFF))==0)
		return 0;
	return a>=0x3F00 ? *p&PALRAMMask:*p;
}

void FCEUPPU_MemPoke(uint16 a,
	uint8 v,
	int hl)
{
	uint8		*p;
(void)hl; /* all writes are high-level */
	if ((p=VAddrSafe(a&=0x3FFF))!=0)
		*p = v;
}

uint8 FCEUPPU_OAMPeek(uint16 a)
{
	return SPRAM[(uint8)a];
}

void FCEUPPU_OAMPoke(uint16 a,
	uint8 v,
	int hl)
{
(void)hl; /* all writes are high-level */
	SPRAM[(uint8)a] = v;
}

int FCEUPPU_MemGetBank(uint16 a)
{ /* CHR bank tracking array not yet implemented */
(void)a;
	return -1;
}

#define LineUpdate()		do{ \
	if (fceuindbg) \
		break; \
	if (PPU_Pline) \
		RefreshLine(LinePixel()); \
} while(0)

STATIC DECLFR(PPUA2002)
{ /* models 1.4, 2 and 3 return the top two bits and a signature */
	uint8		ret;
(void)A;
	LineUpdate();
	ret = (PPU_status&STAT_RDMASK)|(PPUGenLatch&A02LatMask)|A02Version;

	if (fceuindbg)
		return ret;
	vtoggle = 0;
	PPU_status&= STAT_VCLR;

	SetDecayLatch(ret);
	return ret;
}

STATIC DECLFR(PPUA200x)
{
(void)A;
	LineUpdate();

	return PPUGenLatch;
}

STATIC DECLFR(PPUA2004)
{
	uint8		ret;
(void)A;
	LineUpdate();
	ret = 0xFF;
	if (PPUNoFrame) /* no rendering */
	{
		if (PPUSPL<8)
			ret = SPRAM[PPUSPL];
		else if (PPUReg[3]>7)
			ret = SPRAM[PPUReg[3]];
	} else
	{
		ret = OAMLatch;
		A = LinePixel()-257;
		if (A<=320-257u)
			ret = SPRAM[(A>>1&~3)+((A&7)>3 ? 3:(A&7))]; /* should be secondary buf */
	}
	if (!fceuindbg)
		SetDecayLatch(ret);
	return ret;
}

static void VIncPos(uint32 oldA)
{ /* this more elaborate routine fixes "Young Indiana Jones Chronicles" */
	LineUpdate();
	if (PPUNoFrame)
		RefreshAddr = (RefreshAddr+(IncByV ? 32:1))&0x7FFF;
	else if (IncByV)
	{
		if (!(~RefreshAddr&0x7000))
		{ /* HBlank behavior */
			RefreshAddr&= 0xFFF;
			switch(0x3E0)
			{
			case 0x3A0:
				RefreshAddr^= 0x800;
			case 0x3E0:
				RefreshAddr&= 0xFC1F;
				break;

			default:
				RefreshAddr+= 32;
			}
		} else
			RefreshAddr+= 0x1000;
	} else if (!(~RefreshAddr&31))
		RefreshAddr^= 0x41F;
	else
		++RefreshAddr;
	if (PPU_hook)
		PPU_hook(oldA);
}

STATIC DECLFR(PPUA2007)
{ /* updates by 4 cycles, CPU gets feedback on address */
	uint64		t;
	uint32		oldA;
	uint8		ret;

	if (PPUBoot)
		return 0;
	oldA = A = RefreshAddr&POS_MASKS;
	if (!fceuindbg)
		VIncPos(A);
	if (!(~A&0x3F00))
	{ /* palette read */
		int		i;

		ret = (PALRAM[(i=A&31)==0 ? 16:i]&PALRAMMask)|(PPUGenLatch&0xC0);
		A-= 0x1000;
	} else
		ret = VRAMBuffer;
	if (fceuindbg)
		return ret;

	VRAMBuffer = *VAddr(A);

	if (PPU_hook)
		PPU_hook(RefreshAddr&POS_MASKS);
	if (PPU_A12hook && (PPUReg[0]&0x38)!=0x18 && (RefreshAddr&0x1000)>(oldA&0x1000))
		PPU_A12hook();
	/* two reads in a row may stall results */
	t = timestampbase; (void)uint64addu32(t,timestamp);
	if (int64gt(A2007nexttime,t))
		ret = PPUGenLatch;
	(void)uint64addu32(t,3);
	A2007nexttime = t;
	SetDecayLatch(ret);
	return ret;
}

static DECLFR(RGBPPUA2007)
{ /* RGB PPU doesn't support palette reads; also, no warmup */
	uint64		t;
	uint32		oldA = A = RefreshAddr&POS_MASKS;
	uint8		ret;

	if (PPUBoot && FSettings.usercpumodel<5)
		return 0;
	if (!fceuindbg)
		VIncPos(A);
	if (A>0x2000) /* not sure about 3xxx-on addressing */
		A&= 0x2FFF;
	ret = VRAMBuffer;
	if (fceuindbg)
		return ret;
	VRAMBuffer = *VAddr(A);
	if (PPU_hook)
		PPU_hook(RefreshAddr&POS_MASKS);
	if (PPU_A12hook && (PPUReg[0]&0x38)!=0x18 && (RefreshAddr&0x1000)>(oldA&0x1000))
		PPU_A12hook();
	t = timestampbase; (void)uint64addu32(t,timestamp);
	if (int64gt(A2007nexttime,t))
		ret = PPUGenLatch;
	(void)uint64addu32(t,3);
	A2007nexttime = t;
	SetDecayLatch(ret);
	return ret;
}

/* writing at sensitve times resets the PF line address register (not implemented here) */
/* internal flag transfers are delayed by 1.5 PPU cyc (not implemented here) */

STATIC DECLFW(PPUB2000) //hv latch
{
	int		sv;
(void)A;
	if (PPURegBoot)
		return;
	LineUpdate();
	sv = (uint32)V<<HV_SCR_SHIFT;
	TempAddr&= 0xF3FF, TempAddr|= sv&0xC00;

	SetDecayLatch(V);
	V = PPUReg[0];
	PPUReg[0] = PPUGenLatch;
	/* toggling 2000.7 during VBL w/o reading 2002 can cause /NMI to be pulled low multiple times */
	if ((VBlankON&PPU_status)>V) //!VBlankON && V&PPU_status
	{
		//XFrame.NMISR|= 0x80; /* not correct behavior; breaks carts */
		XFrame.intlow|= FCEU_IQNMI2; //TriggerNMI2
	}
}

STATIC DECLFW(PPUB2001) //render
{
(void)A;
	if (PPURegBoot)
		return;
	LineUpdate();
	SetDecayLatch(PPUReg[1]=V);
	FCEUPPU_UpdatePalette();
}

STATIC DECLFW(PPUB2002) //status
{
(void)A;
	SetDecayLatch(V);
}

STATIC DECLFW(PPUB2003) //OAM addr
{
(void)A;
	SetDecayLatch(V);
	PPUReg[3] = V;
	PPUSPL = V&7;
}

DECLFW(PPUB2004) //OAM data
{
(void)A;
	if (PPUNoFrame) /* 2010.05.11, 2010.05.27 */
	{
		if ((PPUReg[3]&3)==2)
			V&= 0xE3; /* some bits aren't stored */
	} else
		V = 0xFF;
	SetDecayLatch(V);

	if (PPUSPL<8)
		SPRAM[PPUSPL] = V;
	else if (PPUReg[3]>7)
		SPRAM[PPUReg[3]] = V;
	++PPUReg[3];
	++PPUSPL;
}

STATIC DECLFW(PPUB2005) //scroll-toggle h/v
{
	if (PPURegBoot)
		return;
	A = TempAddr;
	LineUpdate();
	SetDecayLatch(V);
	if (vtoggle^=1) /* addresses are only 15-bits wide */
		A&= 0x7FE0, A|= V>>3, XOffset = V&7;
	else
		A&= 0xC1F, A|= ((uint32)V<<2|(uint32)V<<12)&0x73E0;
	TempAddr = A;
}

STATIC DECLFW(PPUB2006) //address latch
{
	if (PPURegBoot)
		return;
	LineUpdate();
	A = RefreshAddr;
	SetDecayLatch(V);
	if (vtoggle^=1)
		TempAddr&= 0xFF, TempAddr|= (uint32)V<<8&0x3F00;
	else
	{
		TempAddr&= 0x7F00, RefreshAddr = TempAddr|= V;
		if (PPU_hook)
			PPU_hook(RefreshAddr);
	}
	if (PPU_A12hook && (PPUReg[0]&0x38)!=0x18 && (RefreshAddr&0x1000)>(A&0x1000))
		PPU_A12hook();
}

STATIC DECLFW(PPUB2007) //post-incr addr data write
{
	uint32		k, addrcode = RefreshAddr&POS_MASKS;

	A = RefreshAddr;
	SetDecayLatch(V);
	if (addrcode>=0x3F00)
	{
		uint8	pel;

		if (!RenderON)
			LineUpdate(); /* for full_nes_palette, Micro Machines; 2010.04.13 */

		k = addrcode&0x1F;
		/* bit #6 set so spr pixel won't be zero */
		PALRAM[k] = V, PPUPal[k] = pel = (V&PALRAMMask)|PLB_S;
		if (!(k&3)) /* transparent attr entries are shared */
			k^= 0x10, PALRAM[k] = V, PPUPal[k] = pel;
		/* index #0 for transparent (based on trial and error) */
		PPUPal[0] = PPUPal[4] = PPUPal[8] = PPUPal[12] = (PALRAM[0]&PALRAMMask)|PLB_B;
	} else
	{
		k = addrcode>>HV_SCR_SHIFT;
		if (addrcode<0x2000)
		{
			if ((uint8)(PPUCHRRAM&1u<<k))
				VPage[k][addrcode] = V;
		} else
		{
			k&= 3;
			if ((uint8)(PPUNTARAM&1u<<k))
				vnapage[k][addrcode&0x3FF] = V;
		}
	}

	if (IncByV)
		addrcode+= 0x20;
	else
		++addrcode;
	RefreshAddr = addrcode;
	if (PPU_hook)
		PPU_hook(addrcode&POS_MASKS);
	if (PPU_A12hook && (PPUReg[0]&0x38)!=0x18 && (RefreshAddr&0x1000)>(A&0x1000))
		PPU_A12hook();
}

void FCEUPPU_LineUpdate(void)
{
	LineUpdate();
}

/* (incomplete) OAM emulation */

uint8		PPUsphitdata;
PLB_pel		PPUsprlinebuf[256+8];
int		PPUsphitx;

STATIC void CheckSpriteHit(int e)
{
	const PLB_pel	*r;
	int		x = PPUsphitx;
	uint8		d;

	/* optimized version using Dan Weiss's suggestion; 2010.04.08 */
	if ((e-=16)>254)
		e = 255; /* end before 255; should fix "Battletoads" */
	if ((x+=8)<e) /* take and pin, 8 max (one sprite) */
		e = x;
	x-= 8;
	//if (x<8 && !SpFirst8)
	//	x = 8;
	if (x>=e)
		return;

	r = PPU_Plinef;
	if ((uint8)((d=PPUsphitdata)&0x80) && !(uint8)(r[x]&PLB_B))
		goto hit;
	if (++x>=e)
		return;
	do
	{
		if (!(uint8)((d+=d)&0x80) || (uint8)(r[x]&PLB_B))
			continue;
	hit:	PPU_status|= STAT_OAM_ZH;
		PPUsphitx = 0x100;
		break;
	} while(++x<e);
}

uint8		PPUOAMrefreshed = 0; /* spork the world. */
/*	Any sprites on this line?
	Then this will be set to 1.  Needed for zapper
	emulation and *gasp* sprite emulation.
*/

#define V_FLIP		0x80
#define H_FLIP		0x40
#define SP_BACK		0x20

typedef struct SPR
{
        uint8	y, no, atr, x;
} SPR;

typedef struct SPRB
{
        uint8	ca[2], atr, x;
} SPRB;


uint8		PPUnumsprites, PPUlineHasSp0;

#define SpFetchAddr(A,S,O,V,P) do{A=S->no; \
	if (P) \
		A=(A<<12&0x1000)+(A<<4&0xFE0); \
	else \
		A=(A<<4)+O; \
	if (S->atr&V_FLIP) \
		A+=7,A-=V,A+=P,A-=V&8; \
	else \
		A+=V,A+=V&8; \
}while(0)

static void FetchSpriteData(void)
{
	SPRB		dst;
	SPR		*S = (SPR*)SPRAM;
	uint32		vofs, Pf = Sprite16>>2, H = 8+Pf;
	int		n = 64, ns = 0, ms = PPUmaxsprites;
	uint8		spzonline = 0; /* delay for callbacks */

	vofs = (uint32)((SpAdrHI*2)&~(Pf+=Pf))<<8;
	if (!PPU_hook)
	{
		do
		{
			unsigned	V = scanline-S->y;

			if (V>=H)
				continue;
			if (ns>=ms)
				{PPU_status|= STAT_OAM_OV; break;}
			if (n==64) /*n==63 [n was 0-63 in `for`]*/
				spzonline = 1;

			{
				uint8		*C;
				unsigned	vadr;

				SpFetchAddr(vadr,S,vofs,V,Pf);

				/* Fix this geniestage hack */
				if (MMC5Ex && GenieStage!=1)
					C = MMC5SPRVRAMADR(vadr);
				else
					C = VRAMADR(vadr);

				dst.ca[0] = *C, dst.ca[1] = C[8];
				*(int16*)&dst.atr = *(int16*)&S->atr; /*atr,x*/
				*(int32*)&SPRBUF[ns<<2] = *(int32*)&dst;
			}
			++ns;
		} while(((uint8*)S)+=4,--n);

		if (ns>8) /* handle >8 sprites per scanline */
			PPU_status|= STAT_OAM_OV;
	} else
	{
		do
		{
			unsigned	V = scanline-S->y;

			if (V>=H)
				continue;
			if (ns>=ms)
				{PPU_status|= STAT_OAM_OV; break;}
			if (n==64)
				spzonline = 1;

			{
				uint8		*C;
				unsigned	vadr;

				SpFetchAddr(vadr,S,vofs,V,Pf);

				if (MMC5Ex)
					C = MMC5SPRVRAMADR(vadr);
				else
					C = VRAMADR(vadr);
				dst.ca[0] = *C;

				if (ns<8)
				{
					PPU_hook(0x2000);
					PPU_hook(vadr);
				}

				dst.ca[1] = C[8];
				*(int16*)&dst.atr = *(int16*)&S->atr; /*atr,x*/
				*(int32*)&SPRBUF[ns<<2] = *(int32*)&dst;
			}
			++ns;
		} while(((uint8*)S)+=4,--n);

		if (ns>8)
			PPU_status|= STAT_OAM_OV;
		else if ((n=8-ns)>0)
			do
			{
				PPU_hook(0x2000);
				PPU_hook(vofs);
			} while(--n);
	}

	PPUnumsprites = ns;
	PPUlineHasSp0 = spzonline;
}

#ifndef env68k
STATIC void RefreshSprites(void)
{
	SPRB		*spr;
	int		n;

	PPUOAMrefreshed = 0;
	if ((n=PPUnumsprites)==0)
		return;
	FCEU_dwmemset(PPUsprlinebuf,0,256);
	PPUnumsprites = --n;
	spr = n+(SPRB*)SPRBUF;
	do
	{
		register uint32	pixdata;
		register uint8	J = spr->ca[0], atr;

		PLB_pel	*C, *VB;
		int	x;

		pixdata = PPU_LUT1to4[spr->ca[1]], pixdata+= pixdata;
		pixdata|= PPU_LUT1to4[J];
		if ((J|=spr->ca[1])==0)
			continue;
		x = spr->x, atr = spr->atr;
		if (!n && PPUlineHasSp0 && !(PPU_status&STAT_OAM_ZH))
		{
			if (!(uint8)(atr&H_FLIP))
				PPUsphitdata = J;
			else
				PPUsphitdata = bitreverseLUT[J];
			PPUsphitx = x;
		}

		C = PPUsprlinebuf+x;
		VB = (PPUPal+0x10)+((atr&3)<<2);

		if ((uint8)(atr&SP_BACK))
			if ((uint8)(atr&H_FLIP))
			{
				if ((uint8)(J&0x80))
					C[7] = VB[pixdata&3];
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[6] = VB[pixdata&3];
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[5] = VB[pixdata&3];
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[4] = VB[pixdata&3];
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[3] = VB[pixdata&3];
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[2] = VB[pixdata&3];
				if ((J+=J)&0x80)
					C[1] = VB[pixdata>>4&3];
				if ((J+=J)&0x80)
					*C = VB[pixdata>>=8];
			} else
			{
				if ((uint8)(J&0x80))
					*C = VB[pixdata&3];
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[1] = VB[pixdata&3];
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[2] = VB[pixdata&3];
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[3] = VB[pixdata&3];
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[4] = VB[pixdata&3];
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[5] = VB[pixdata&3];
				if ((J+=J)&0x80)
					C[6] = VB[pixdata>>4&3];
				if ((J+=J)&0x80)
					C[7] = VB[pixdata>>=8];
			}
		else
			if ((uint8)(atr&H_FLIP))
			{
				if ((uint8)(J&0x80))
					C[7] = VB[pixdata&3]|PLB_B;
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[6] = VB[pixdata&3]|PLB_B;
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[5] = VB[pixdata&3]|PLB_B;
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[4] = VB[pixdata&3]|PLB_B;
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[3] = VB[pixdata&3]|PLB_B;
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[2] = VB[pixdata&3]|PLB_B;
				if ((J+=J)&0x80)
					C[1] = VB[pixdata>>4&3]|PLB_B;
				if ((J+=J)&0x80)
					*C = VB[pixdata>>=8]|PLB_B;
			} else
			{
				if ((uint8)(J&0x80))
					*C = VB[pixdata&3]|PLB_B;
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[1] = VB[pixdata&3]|PLB_B;
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[2] = VB[pixdata&3]|PLB_B;
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[3] = VB[pixdata&3]|PLB_B;
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[4] = VB[pixdata&3]|PLB_B;
				pixdata>>= 4;
				if ((J+=J)&0x80)
					C[5] = VB[pixdata&3]|PLB_B;
				if ((J+=J)&0x80)
					C[6] = VB[pixdata>>4&3]|PLB_B;
				if ((J+=J)&0x80)
					C[7] = VB[pixdata>>=8]|PLB_B;
			}
	} while(--spr,--n>=0);

	PPUlineHasSp0 = 0;
	PPUOAMrefreshed = 1;
}
#endif

static void CopySprites(PLB_pel *target)
{
	const PLB_pel	*S; /* sprite/OAM */
	PLB_pel		*P; /* playfield */
	uint8		n;

	if (!PPUOAMrefreshed)
		return;
	n =~PPUReg[1], n+= n, n&= 8; /* have SpFirst8 determine where to start */
	S = n+PPUsprlinebuf;
	P = n+target;

	if (PPUrendis&1) /* sprites hidden */
		goto end;
	if (PPUrendis&4) /* playfield off (pixel data could be anything) */
	{
		do
		{
			PLB_pel		t;

			if (!*(uint32*)S)
				{S+= 4; continue;}
			if ((t=*S++)!=0)
				t|=PLB_B, *P = t; /* background priority should be set */
			if ((t=*S++)!=0)
				t|=PLB_B, P[1] = t;
			if ((t=*S++)!=0)
				t|=PLB_B, P[2] = t;
			if ((t=*S++)!=0)
				t|=PLB_B, P[3] = t;
		} while(P+=4,n+=4);
		goto end;
	}

	do
	{
		PLB_pel		t;

		if (!*(uint32*)S)
			{S+= 4; continue;}
	/*act on each normal or behind bg sprite pixel*/
		if ((t=*S++)!=0 && (uint8)((t|*P)&PLB_B))
			*P = t;
		if ((t=*S++)!=0 && (uint8)((t|P[1])&PLB_B))
			P[1] = t;
		if ((t=*S++)!=0 && (uint8)((t|P[2])&PLB_B))
			P[2] = t;
		if ((t=*S++)!=0 && (uint8)((t|P[3])&PLB_B))
			P[3] = t;
	} while(P+=4,n+=4);
end:	PPUOAMrefreshed = 0;
}

/* line rendering */

static void DoHBlank(void)
{
	uint32		rad, t;

	if (!RenderON)
		return;
	rad = RefreshAddr;
	if (!(~rad&0x7000)) /*(rad&0x7000)==0x7000*/
		if ((t=~(rad^=0x7000)&0x3E0)==0) /*(rad&0x3E0)==0x3E0*/
			rad^= 0x3E0;
		else if (t==0x40) /*(rad&0x3E0)==0x3A0*/
			rad^= 0x3A0^0x800;
		else
			rad+= 0x20;
	else
		rad+= 0x1000;
	RefreshAddr = rad;
}

static void blankPixels(uint8 c,
	int z,
	int m)
{
	PLB_pel		*d;
	uint32		t, x = linepel-MIN(linepel,m);

	if (z>256)
		z = 256;
	if ((z-=x)<0)
		return;
	d = PPU_Plinef+x, t = exp8(c);
	if (z&1)
		*d++ = t;
	if (z&2)
		*((uint16*)d)++ = t;
	if (z&4)
		*((uint32*)d)++ = t;
	if (z&8)
		*((uint32*)d)++ = t,*((uint32*)d)++ = t;
	if (z&16)
		*((uint32*)d)++ = t, *((uint32*)d)++ = t,
			*((uint32*)d)++ = t,*((uint32*)d)++ = t;
	if (z&=-32)
		do
			*((uint32*)d)++ = t, *((uint32*)d)++ = t,
				*((uint32*)d)++ = t, *((uint32*)d)++ = t,
				*((uint32*)d)++ = t, *((uint32*)d)++ = t,
				*((uint32*)d)++ = t, *((uint32*)d)++ = t;
		while(z+=-32);
}

static void FASTAPASS(1) RefreshLine(int lastpixel)
{
	register PLB_pel *P;

	uint32		smorkus = RefreshAddr;
#define RefreshAddr smorkus
	uint32		vofs;
	int		X1, pfetchtile, pfetchpel;
	int		numtiles, firsttile;
#ifndef env68k
	int		xdownshift = 8-XOffset;
#endif
	int		scantileline = scanline>>3;

	/* Yeah, recursion would be bad.
		PPU_hook() functions can call mirroring/chr bank switching functions,
		which call FCEUPPU_LineUpdate, which calls this function
	*/
	static uint8 norecurse = 0;
	if (norecurse)
		return;
	pfetchpel = lastpixel;
#if 1
	if (PPUsphitx!=0x100 && !(PPU_status&STAT_OAM_ZH))
		if (PPUsphitx<lastpixel-16
				&& PPUsphitx>=(pfetchpel&~7)-16)
			pfetchpel+= 8; /*first 8 must be rendered*/
#endif
	if (pfetchpel>272)
		pfetchpel = 272;
	pfetchtile = pfetchpel>>3;
	firsttile = linepel>>3;
	if ((numtiles=pfetchtile-firsttile)<=0)
		return;
	scantileline = scanline>>3;
	P = PPU_Pline;

	if (!RenderON)
	{ /* live color */
		if ((X1=RefreshAddr-0x3F00)+0u>255u)
			X1 = 0;
		numtiles = numtiles*8u;
		/* don't use PPUPal */
		blankPixels((PALRAM[X1&63]&PALRAMMask)|PLB_B,pfetchpel,23);
		P+= numtiles;

#define TOFIXNUM (272-4)
		if (lastpixel>=TOFIXNUM)
			linereset = 0; /* not rendering, no HBlank addr update */
		goto end;
	}

	vofs = (uint32)BGAdrHI<<8|(RefreshAddr>>12&7);

	if (PPUrendis&4) /* skip playfield rendering (yes, some code put here) */
	{
		uint32		me32;
		//substitute void color
		me32 = exp8(PPUvoidcolor!=255 ? PPUvoidcolor:PPUPal[4]);
		for(X1=firsttile; X1<pfetchtile; ++X1)
		{ /* fake everytin' */
			if (PPU_hook)
			{
				PPU_hook(0x2000|(RefreshAddr&0xFFF));
				PPU_hook(vofs);
			}

			if (pfetchtile<32)
				*((uint32*)P)++ = me32, *((uint32*)P)++ = me32;

			if (!(~RefreshAddr&HT_POS_MASK))
				RefreshAddr^= H_POS_MASKS;
			else
				++RefreshAddr;

			if (PPU_hook)
				PPU_hook(0x2000|(RefreshAddr&0xFFF));
		}
		atlatch = 0;
		goto skip;
	}
	/* This high-level graphics MMC5 emulation code was written
		for MMC5 carts in "CL" mode.  It's probably not totally
		correct for carts in "SL" mode.
	*/
#define PPUT_MMC5
	if (MMC5Ex && GenieStage!=1)
	{
		if (!MMC5ExMode && MMC5SplitCtrl&0x80)
		{
			int	tochange = (MMC5SplitCtrl&31)-firsttile;

			for(X1=firsttile; X1<pfetchtile; ++X1)
				if ((--tochange<0 && MMC5SplitCtrl&0x40)
						|| (tochange>=0 && !(MMC5SplitCtrl&0x40)))
				{
				#define PPUT_MMC5SP
				#include "PPUtile.h"
				#undef PPUT_MMC5SP
				} else
				{
				#include "PPUtile.h"
				}
		} else if (MMC5ExMode==1 && MMC5SplitCtrl&0x80)
		{
			#define PPUT_MMC5SP
			#define PPUT_MMC5CHR1
			for(X1=firsttile; X1<pfetchtile; ++X1)
			{
			#include "PPUtile.h"
			}
			#undef PPUT_MMC5CHR1
			#undef PPUT_MMC5SP
		} else if (MMC5ExMode==1)
		{
			#define PPUT_MMC5CHR1
			for(X1=firsttile; X1<pfetchtile; ++X1)
			{
			#include "PPUtile.h"
			}
			#undef PPUT_MMC5CHR1
		} else
			for(X1=firsttile; X1<pfetchtile; ++X1)
			{
			#include "PPUtile.h"
			}
	#undef PPUT_MMC5
	} else if (PPU_hook)
	{
		norecurse = 1;
		#define PPUT_HOOK
		for(X1=firsttile; X1<pfetchtile; ++X1)
		{
		#include "PPUtile.h"
		}
		#undef PPUT_HOOK
		norecurse = 0;
	} else
		for(X1=firsttile; X1<pfetchtile; ++X1)
		{
		#include "PPUtile.h"
		}
#undef RefreshAddr
skip:	RefreshAddr = smorkus;
	if (!BGFirst8 && PPU_Pline==PPU_Plinef) /*firsttile<3 && pfetchtile>2*/
	{
		uint32		*d = (uint32*)PPU_Plinef;

		*d = *d++ = exp8(PPUPal[0]);
	}

	if (!PlayfdON)
		blankPixels(PPUPal[0],pfetchpel,23); /* 23 currently aligns properly */

	if (lastpixel>=TOFIXNUM && linereset)
	{
		DoHBlank();
		linereset = 0;
	}

	CheckSpriteHit(lastpixel);

end:	if (InputScanlineHook && lastpixel>=16)
		InputScanlineHook(PPU_Plinef,
			PPUOAMrefreshed ? PPUsprlinebuf:0,
			linestartts,
			pfetchtile*8u-16);

	PPU_Pline = P;
	linepel = pfetchpel;
}


#define DoUpdateV()	do{ \
	if(RenderON) \
		RefreshAddr=(RefreshAddr&0xFBE0)|(TempAddr&0x41F); \
}while(0)

#define DoCYC320_5()	do{ \
/*320*/	OAMLatch=SPRAM[0]; X65xx_Run(CPUSC[PPU_5]); \
}while(0)

static void ResetRL(void)
{
//	FCEU_memsetl(target,-1,256); -- no longer necessary at all; 2011.02.14
	if (InputScanlineHook)
		InputScanlineHook(0,0,0,0);
	PPU_Pline = PPU_Plinef;
	linepel = 0;
	linestartts = timestamp*CPU_SUBCYC+XFrame.count;
	linereset = 0;
	LineUpdate();
	linereset = 1;
}

static void DoLine(PLB_pel *target)
{
/*	{ // needs to be grabbed before line, better rendering altogether...
		int		t;

		if ((t=RefreshAddr-0x3F00)+0u>255u)
			t = 0;
		PPU_PStart[scanline] = epal_inv[PPUEmphasis+(PALRAM[t&63]&PALRAMMask)];
	}
*/
	if (MMC5Ex && RenderON)
	{ /* fixes alignment - does it break anything?; 2015.08.22 */
	//	X6502_Run(10);
		MMC5_hb(scanline);
	//	X6502_Run(64-10);
	}// else
	//	X65xx_Run(CPUSC[PPU_64]);

/*
 Visual2C02:
 0  idle
 1  load NT byte
 2  transfer
 3  load AT byte
 4  transfer
 5  low PF tile byte
 6  transfer
 7  hight PF tile byte
 8  transfer, increment horiz.
..
257 horiz./vert.
*/

/* 2004 returns from secondary OAM */
/* secondary is written with values read from primary */

/*0*/	OAMLatch = 0xFF;
	X65xx_Run(CPUSC[PPU_64]);
/*64*/	PPUEmphasis = EmphLUT[PPUReg[1]>>5];
	OAMLatch = 0; X65xx_Run(CPUSC[PPU_192]);
/*256*/	OAMLatch = 0xFF;

	/* EndRL */
	RefreshLine(272);
	if (linereset)
		DoHBlank();
	CheckSpriteHit(272); /* finish */
	PPU_Pline = 0;

	if (PPUrendis&2) /* User asked to not display background data. */
	{
		uint32		t = exp8(PPUvoidcolor!=255 ? PPUvoidcolor:PPUPal[0]);

		{
			uint32		*d = (uint32*)PPU_Plinef, c = 256/32;

			do
				*d++ = *d++ = *d++ = *d++ =
				*d++ = *d++ = *d++ = *d++ = t;
			while(--c);
		}
	}

	if (SpriteON) /* this should be in RefreshLine */
		CopySprites(PPU_Plinef);

	if (PPUrendis&4)
	{ /* SPZ hit'll fail without something; all have bit #7 on; 2010.04.26 */
		if (SpriteON && scanline==SPRAM[0]) /* only if SpriteON; 2010.05.27 */
			PPU_status|= STAT_OAM_ZH;
	}
//256
	PPUsphitx = 0x100;

	if (RenderON)
		FetchSpriteData();

	if (GameHBIRQHook && RenderON && (PPUReg[0]&0x38)!=0x18)
	{
		X65xx_Run(CPUSC[PPU_6]);
		DoUpdateV(); //262
		X65xx_Run(CPUSC[PPU_4]);

		GameHBIRQHook();

		X65xx_Run(CPUSC[PPU_54]); /* 85-16-10-5 */
		DoCYC320_5();
	} else
	{
		X65xx_Run(CPUSC[PPU_6]);
		DoUpdateV(); //262
		X65xx_Run(CPUSC[PPU_58]); /* 85-16-6-5 */
		DoCYC320_5();

		// A semi-hack for Star Trek: 25th Anniversary
		if (GameHBIRQHook && RenderON && (PPUReg[0]&0x38)!=0x18)
			GameHBIRQHook();
	}
//325
	if (SpriteON)
		RefreshSprites();
	if (GameHBIRQHook2 && RenderON)
		GameHBIRQHook2();
	/* remap the line with emphasis after all the hooks */
	_memrem8(target,PPU_Plinef,256,epal_inv+PPUEmphasis);

	ResetRL();
	//X65xx_Run(CPUSC[PPU_4]);
	//NextFirst8 = SpFirst8;
	//X65xx_Run(CPUSC[PPU_12]);
	X65xx_Run(CPUSC[PPU_16]);
}

static void BuildEmphLUT(int which)
{
	uint32		c = 7;

	do
		EmphLUT[c] = c<<8;
	while(!(--c&0x80000000u));

	if (which) /* PAL or Dendy: swap green and red */
		EmphLUT[1] = 2<<8, EmphLUT[2] = 1<<8, EmphLUT[5] = 6<<8, EmphLUT[6] = 5<<8;
}

void FCEUPPU_SetVideoSystem(int which)
{
	static uint16	t[] = {1, 2, 3, 4, 5, 6, 12, 16, 54, 58, 64, 85, 192, 256, LCYCS};

	unsigned	i, fact;

	FSettings.FirstSLine = FSettings.UsrFirstSLine[1];
	FSettings.LastSLine = FSettings.UsrLastSLine[1];
	if (which==GIV_PAL)
	{ /* authentic PAL */
		fact = PAL_SC;
		CPUSC[PPU_VBL] = PAL_SC*LCYCS*FSettings.UsrVBLSLines[1]; /*312-242*/
		CPUSC[PPU_odd] = PAL_SC*4;
		CPUSC[PPU_EFSleep] = PAL_SC*(LCYCS-1);
		CPUSC[PPU_BootSync] = PAL_SC*(BOOTSYNCCYC-1);
		CPUSC[PPU_visible] = PAL_SC*LCYCS*240;
	} else
	{ /* NTSC, RGB and clones */
		fact = NTSC_SC;
		if (which==GIV_NTSC || which==GIV_RGB)
		{ /* except for the palette, RGB follows NTSC conventions */
			FSettings.FirstSLine = FSettings.UsrFirstSLine[0];
			FSettings.LastSLine = FSettings.UsrLastSLine[0];
			CPUSC[PPU_odd] = NTSC_SC*3;
		} else
			CPUSC[PPU_odd] = NTSC_SC*4; /* no odd-drop on Dendy, apparently */
		CPUSC[PPU_VBL] = NTSC_SC*LCYCS*FSettings.UsrVBLSLines[0]; /*262-242*/
		CPUSC[PPU_EFSleep] = which==GIV_Dendy ? NTSC_SC*(51*LCYCS-1):NTSC_SC*(LCYCS-1);
		CPUSC[PPU_BootSync] = NTSC_SC*(BOOTSYNCCYC-1);
		CPUSC[PPU_visible] = NTSC_SC*LCYCS*240;
	}
	i = PPU_1;
	do
		CPUSC[i] = fact*(t-1)[i];
	while(++i<PPU_odd);

	if ((CPUSC[PPU_VBL]-=CPU_NMI_SC)<0)
		CPUSC[PPU_VBL] = 0; /* idiot proofing */
	/* NMI_DELAY contained CPU cycles in NTSC PPU cycles (incorrect PAL timing) */
	CPUSC[PPU_NMI_M2] = CPU_NMI_SC-NMI_DISABLE_DELAY;

	XBuf = FCEU_video+(((PPUVFrameCounter^PPUVFirstField)&1)*256u);

	BuildEmphLUT(which);

	/* chip functions may change by user */
	{
		readfunc	*rdrt, r04, r07;
		int		x;

		OptFAddr(r04,FSettings.usercpumodel>4 ? PPUA2004:PPUA200x);
		OptFAddr(r07,PPUChipType&(RP2C03B|RP2C03G)
			|| FSettings.uservidsys==GIV_RGB
			|| FSettings.usercpumodel<5 ? RGBPPUA2007:PPUA2007);
		x = 0x2000, rdrt = x+ARead;
		do
		{
			rdrt[4] = r04;
			rdrt[7] = r07;
		} while((x+=8)<0x4000);
	}
}

void FCEUPPU_Init(void)
{
	makeluts(bitreverseLUT
#ifndef env68k
		,PPU_LUT1to4
		,PPU_ATLUT
#endif
		);
	BuildEmphLUT(0);
}

void FCEUPPU_Reset(int NESboot)
{
	XBuf = FCEU_video+((PPUVFirstField&1)*256u); /* odd field when frame begins */
	VRAMBuffer = PPUReg[0] = PPUReg[1] = PPUReg[3] = PPUSPL =
		PPUGenLatch = vtoggle = 0;
	RefreshAddr = TempAddr = PPUVFrameCounter = 0;
	PPUBoot = 2;
	PPURegBoot = NESboot ? 3:0;
	_int64xset(A2007nexttime,0);

	A02Version = 0, A02LatMask = 0x1F, PALRAMMask = 0x3F;

	if (PPUChipType&RC2C05_04) /* 2C05-01-04 */
		A02Version = 0x1B, A02LatMask = 0;
	else if (PPUChipType&RC2C05_03)
		A02Version = 0x1C, A02LatMask = 0;
	else if (PPUChipType&RC2C05_02)
		A02Version = 0x3D, A02LatMask = 0;
}

/* from simple memset of zero...; 2010.04.14, 2010.05.26 */
static void InitializePalette(uint8 *d)
{ /* NOTE: there are multiple variants of the powerup palette;
     This table is for 'blargg_ppu_tests_2005.09.15b/power_up_palette.nes';
    powerup palette isn't always solid, as 2007 reads can reflect different results.

     Index #0 is 9 (the yellow-green colorburst color) at boot.
     All by-attribute transparent entries are shared (28 bytes of palette RAM);
    however, the render-off color will vary until manually set.
  */
	static const uint8	t[32] =
	{ /* uninitialized #0 set to black */
	/*+00*/	0x3F,0x01,0x00,0x01,0x00,0x02,0x02,0x0D, 0x08,0x10,0x08,0x24,0x00,0x00,0x04,0x2C,
	/*+10*/	0x09,0x01,0x34,0x03,0x00,0x04,0x00,0x14, 0x08,0x3A,0x00,0x02,0x00,0x20,0x2C,0x08
	};
	FCEU_memcpy(d,t,32);
}

void FCEUPPU_UpdatePalette(void)
{
	uint32		k = ((GrayLine-1)|0x30303030)&0x3F3F3F3F, a;

	PALRAMMask = k;
	*(int32*)PPUPal = a = (*(int32*)PALRAM&k)|_le32_native(PLB_B);
	k&= _le32_native(~0xFF), a&= _le32_native(0xFF); /* duplicate #0 to all transparent */
	((int32*)PPUPal)[1] = (((int32*)PALRAM)[1]&k)|a;
	((int32*)PPUPal)[2] = (((int32*)PALRAM)[2]&k)|a;
	((int32*)PPUPal)[3] = (((int32*)PALRAM)[3]&k)|a;
	a|= _be32_native(PLB_S*0x10101); /* sprite flag for non-transparent second half */
	((int32*)PPUPal)[4] = (((int32*)PALRAM)[4]&k)|a;
	((int32*)PPUPal)[5] = (((int32*)PALRAM)[5]&k)|a;
	((int32*)PPUPal)[6] = (((int32*)PALRAM)[6]&k)|a;
	((int32*)PPUPal)[7] = (((int32*)PALRAM)[7]&k)|a;
}

void FCEUPPU_Power(void)
{
	readfunc	*rdrt, r0x, r02, r04, r07;
	writefunc	*wrrt, w[8];
	int		x;

	InitializePalette(PALRAM);
	FCEUPPU_UpdatePalette();

	dwmemset(NTARAM,0,0x800);
	dwmemset(SPRAM,~0,0x100); /* FCEU did 0 */
	OAMLatch = 0xFF;
	FCEUPPU_Reset(0); /* no write delays */
	PPU_status = STAT_VINT*0x11>>4;

	if (FCEUGameInfo->type==GIT_NSF)
		return;
	OptFAddr(r0x,PPUA200x);
	OptFAddr(r02,PPUA2002);
	/* early 2C02s dont implement OAM reading (at least adequately) */
	OptFAddr(r04,FSettings.usercpumodel>4 ? PPUA2004:PPUA200x);
	/* apparently, they also don't do palette reads */
	OptFAddr(r07,PPUChipType&(RP2C03B|RP2C03G)
		|| FSettings.uservidsys==GIV_RGB
		|| FSettings.usercpumodel<5 ? RGBPPUA2007:PPUA2007);
	/* RC2C05 chips address 2000 and 2001 writes in reverse */
	/* the mapper change and scrambled palettes were deliberate to curb ROM copying */
	if (PPUChipType&(RC2C05_01|RC2C05_02|RC2C05_03|RC2C05_04|RC2C05_05))
		{OptFAddr(w[0],PPUB2001); OptFAddr(w[1],PPUB2000);}
	else
		{OptFAddr(w[0],PPUB2000); OptFAddr(w[1],PPUB2001);}
	OptFAddr(w[2],PPUB2002);
	OptFAddr(w[3],PPUB2003);
	OptFAddr(w[4],PPUB2004);
	OptFAddr(w[5],PPUB2005);
	OptFAddr(w[6],PPUB2006);
	OptFAddr(w[7],PPUB2007);

	x = 0x2000, rdrt = x+ARead, wrrt = x+BWrite;
	do
		*rdrt++ = r0x,
		*rdrt++ = r0x,
		*rdrt++ = r02,
		*rdrt++ = r0x,
		*rdrt++ = r04,
		*rdrt++ = r0x,
		*rdrt++ = r0x,
		*rdrt++ = r07,

		*wrrt++ = w[0],
		*wrrt++ = w[1],
		*wrrt++ = w[2],
		*wrrt++ = w[3],
		*wrrt++ = w[4],
		*wrrt++ = w[5],
		*wrrt++ = w[6],
		*wrrt++ = w[7];
	while((x+=8)<0x4000);
}

/* PPU-oriented frame */

int FCEUPPU_Loop(int skip)
{
	if (PPUBoot)
	{ /* needed for Knight Rider, Ironsword, others */
		dwmemset(FCEU_video,exp8(PPUvoidcolor),FCEUSCRN_ROWSIZE*240);
	//	dwmemset(PPU_PStart,exp8(PPUvoidcolor),sizeof(PPU_PStart));
		if (PPUBoot>1)
		{
			PPUBoot = 1;
			//PPU_status = STAT_VINT; /* should take last status */
			scanline = 241;
			X65xx_Run(CPUSC[PPU_BootSync]);
			PPU_status = 0;
			scanline = 0;
			X65xx_Run(CPUSC[PPU_visible]);
			goto done;
		}
		/* second boot frame is where things become operable */
		/* wakeup to NMI is wildly inconsistent with a real RGB PPU */
		PPU_status = STAT_VINT;
		scanline = 241;
		X65xx_Run(CPUSC[PPU_NMI_M2]+CPUSC[PPU_VBL]);
		/* can read-modify here */
		PPUBoot = PPU_status = 0;
		scanline = 0;
		X65xx_Run(CPUSC[PPU_2]);
		/* can write 240+VBL lines + 314 cycles in */
		PPURegBoot = 0;
		X65xx_Run(CPUSC[PPU_visible]);
		goto sleep; /* yeah, I should */
	}
	XBuf = (PLB_pel*)(256^(long int)XBuf); /* swap fields; 2015.08.03 */

	if (PPULatDecay && --PPULatDecay)
		PPUGenLatch = 0;
//340	/* entering vertical interrupt phase */
	PPU_status|= STAT_VINT>>4;
	XFrame.NMI = VBlankON; /* NMI-suppression point */
	X65xx_Run(CPUSC[PPU_1]); /* RGB PPU may be a little early (it fails 5.nmi_suppression) */
	XFrame.NMI&= VBlankON;
	PPU_status = (PPU_status&STAT_VCLR)|(PPU_status<<4&STAT_VINT);

	if (FCEUGameInfo->type==GIT_NSF)
	{
		X65xx_Run(CPU_NMI_SC);
		DoNSFFrame();
	} else
	{
		scanline = 241; /* necessary for 2004 reads; 2009.12.29, 2010.05.29 */

		/* reset sprites */
		PPUReg[3] = PPUSPL = 0; //Not having this here breaks a Super Donkey Kong game.
		X65xx_Run(NMI_DISABLE_DELAY);
		/* according to the documentation, VBL line is a NAND gate of 2000.7 and 2002.7 */
		/* NMI execution timing needs to be correct for Test Cartridge to work properly */
		/* some games with basic 2002 loops will hang if the NMI thwarts the VBL flag */
		PPU_status|= (VBlankON&PPU_status)>>5;
		X65xx_Run(CPUSC[PPU_NMI_M2]); /* NMI_DELAY-1-1 */
		XFrame.NMI = 0;
		if (PPU_status&4) /* NMI needs to be delayed if an ISR was just executed */
			XFrame.intlow|= XFrame.BP&B_FLAG ? FCEU_IQNMI2:FCEU_IQNMI;
	}
	X65xx_Run(CPUSC[PPU_VBL]);

	PPU_status = 0; /* out of VINT; on to no-render evaluation line */
//0
	/* for OAM CYC 1-64, primary read is overridden with FF, erasing secondary */
	OAMLatch = 0xFF;
	X65xx_Run(CPUSC[PPU_256]);

	if (RenderON)
	{
		if (GameHBIRQHook && (PPUReg[0]&0x38)!=0x18)
			GameHBIRQHook();
		if (PPU_hook)
		{
			int	x = 42;

			do
			{
				PPU_hook(0x2000);
				PPU_hook(0);
			} while(--x);
		}
		if (GameHBIRQHook2)
			GameHBIRQHook2();
	}
//256
	X65xx_Run(CPUSC[PPU_64]); /* 85-16-5 */
//320 - the hook should be here
	DoCYC320_5();
//325
	if (RenderON)
	{
		RefreshAddr = TempAddr;
		if (PPU_hook)
			PPU_hook(RefreshAddr&0x3FFF);
	}

	//Clean this stuff up later.
	PPUOAMrefreshed = PPUnumsprites = 0;

	/* frame toggle assessment is done at CYC 338; 2010.04.15 */
	ResetRL();
	X65xx_Run(CPUSC[PPU_12]);
	/* RGB PPU timing differs here */
/* 338 - one less if NTSC, odd and rendering */
	{
		const static int	eoix[2] = {PPU_odd, PPU_even};

		int		even = (PPUVFrameCounter|!RenderON)&1;

		PPUBurstCount = PPUBurstCount+2-even;
		if (PPUBurstCount>2)
			PPUBurstCount-= 3;
		X65xx_Run(CPUSC[eoix[even]]);
	}

	if (FCEUGameInfo->type==GIT_NSF)
		X65xx_Run(PAL ? PAL_SC*NSFCycles:NTSC_SC*NSFCycles);
	else if (skip)
	{
		int	y = 1+SPRAM[0];

		PPU_status|= STAT_OAM_OV; //set overflow- Fixes "Bee 52".  Does it break anything?
		scanline = 0, PPU_Pline = 0;
		if (GameHBIRQHook)
		{
			X65xx_Run(CPUSC[PPU_256]);
			do
			{
				if (RenderON)
					GameHBIRQHook();
				if (scanline==y && SpriteON)
					PPU_status|= STAT_OAM_ZH;
				X65xx_Run(CPUSC[PPU_SL]);
			} while(++scanline<239);
			if (RenderON)
				GameHBIRQHook();
			if (scanline==y && SpriteON)
				PPU_status|= STAT_OAM_ZH;
			X65xx_Run(CPUSC[PPU_85]);
		} else
		{
			if (y>239)
				y = 240;
			else
			{
				for(; scanline<y; ++scanline)
					X65xx_Run(CPUSC[PPU_SL]);
				if (SpriteON)
					PPU_status|= STAT_OAM_ZH;
				y =-y+240;
			}
			do
				X65xx_Run(CPUSC[PPU_SL]);
			while(++scanline,--y);
		}
	} else
	{
		scanline = 0;
		do
			DoLine(XBuf+(scanline*FCEUSCRN_ROWSIZE));
		while(++scanline<240);

		if (MMC5Ex && RenderON)
			MMC5_hb(scanline);
	}
sleep:	X65xx_Run(CPUSC[PPU_EFSleep]);
done:	FCEU_PutImage(PPUVFrameCounter++,skip);
	return skip;
}

static uint16	TempAddrT, RefreshAddrT;

void FCEUPPU_LoadState(int v)
{
	if (v<9817) /* FCE Ultra didn't have PPUPal, only PALRAM */
		PALRAM[4] = PALRAM[8] = PALRAM[12] = PALRAM[0];
	if (v<9822)
		_int64xset(A2007nexttime,0);
	FCEUPPU_UpdatePalette();
	vtoggle&= 1;
	TempAddr = TempAddrT;
	RefreshAddr = RefreshAddrT;
}

SFORMAT FCEUPPU_STATEINFO[] =
{
	{NTARAM, 0x800, "NTAR"},
	{PALRAM, 0x20, "PRAM"},
	{SPRAM, 0x100, "SPRA"},
	{PPUReg, 4, "PPUR"},
	{&PPUBoot, 1, "DEAD"},
	{&PPUSPL, 1, "PSPL"},
	{BIG_ENDIAN*3+&XOffset, 1, "XOFF"},
	{&vtoggle, 1, "VTOG"},
	{&RefreshAddrT, 2|FCEUSTATE_RLSB, "RADD"},
	{&TempAddrT, 2|FCEUSTATE_RLSB, "TADD"},
	{&VRAMBuffer, 1, "VBUF"},
	{&PPUGenLatch, 1, "PGEN"},
	{&PPUVFrameCounter, 4|FCEUSTATE_RLSB, "VFMC"}, /* sound.c uses "FCNT" */
	{&A2007nexttime, 8|FCEUSTATE_RLSB, "A07T"},
	{&PPUBurstCount, 1, "VNBC"},
	{0}
};

void FCEUPPU_SaveState(void)
{
	TempAddrT = TempAddr;
	RefreshAddrT = RefreshAddr;
}
