/* Displays and saves NTSC filtered image. Mouse controls sharpness and gamma.
Defaults to using "test.bmp" for input and "filtered.bmp" for output. Input
image must be an uncompressed BMP. Also writes "nes.pal" RGB color file on exit.

Usage: demo [in.bmp [out.bmp]]

Space   Toggle field merging
C       Composite video quality
S       S-video quality
R       RGB video quality
M       Monochrome video quality
D       Toggle between standard and Sony decoder matrix
*/

#include "nes_ntsc.h"

/* * * from "demo_impl.h" * * */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#define HAVE_OS
#include <MacWindows.h>
#include <QDOffscreen.h>
#include <ImageCompression.h>
#include <Timer.h>
#include "enviro64.h"
//#include <windows.h>
//#include "mwind.h"
//#include "mevent.h"
//#include "mimg_file.h"
#include "nes_ntsc_impl.h"

/* Image loader */
typedef struct image_t
{
	unsigned char const* byte_pixels;/* 8-bit pixels */
	nes_ntsc_out_t const* rgb_out;    /* 16-bit pixels */
	int width;
	int height;
	int row_width; /* number of pixels to get to next row (may be greater than width) */
} image_t;
/* if no palette, loads as 16-bit RGB */
void load_bmp( image_t* out, const char* path, uint8 palette [256][3] );
void save_bmp( const char* path );

int read_input( void );
void lock_pixels( void );
void double_output_height( void );
void display_output( void );
void fatal_error( const char* str );

static unsigned char* output_pixels; /* RGB */
static long output_pitch;
static float mouse_x, mouse_y; /* -1.0 to 1.0 */
static int32 last_mouse;
static int mouse_moved;
static int key_pressed;

/* implementation */

WindowPtr screen_desktop_window;
CGrafPtr screen_desktop_port;
static Rect rect;
static UnsignedWide next_time;

/* ---- required suite */

int8		temporary[512];

void global_idle_proc(void)
{
}

void stop_interface_fade(void)
{
}

uint8 *getpstr(void *d,
	short ID,
	int x)
{
(void)x;
	if (!d || ID<0)
		return NULL;
	*(int8*)d = 0;
	return d;
}

static uint8	_dbgTemp[1024];

#define ATTASKLEVEL0()		1

int dprintf(const char *form,
	...)
{
	va_list		args;
	char		*s;
	int		len;

	va_start(args,form);

	if ((len=vsprintf(1+(char*)_dbgTemp,form,args))>255)
		len = 255;

	va_end(args);

	s = memchr(1+_dbgTemp,';',*_dbgTemp=len);
	//_dbgTemp[len+1] = 0; -- already set by vsprintf

	/* if first ';' follows "g;" */
	if (!ATTASKLEVEL0() || !s || s[1]!='g') // || s[2]!=';')
	{
#if TARGET_OS_MAC
		DebugStr(_dbgTemp);
#else
		__debugbreak();
#endif
	}
	if (s)
		*s = 0;
	if (ATTASKLEVEL0())
		printf("## %s\n",
			1+_dbgTemp);
	return len;
}

#define fatal_error vhalt
# define vhalt(x)			_debugMsg(x,__FILE__,__LINE__|3<<30)

void _debugMsg(const char *msg,
	const char *where,
	int n)
{
	static char	*type[] = {"warning","assertion failure","pause","fatal error"};
	int		z, i = n&0x3FFFFFFF;

	n = (uint32)n>>30;
	z = sprintf((char*)_dbgTemp+1,msg ? "%s in '%s', line %d: %s\n":"%s in '%s', line %d\n",
		type[n],where,i,msg);
	*_dbgTemp = MIN(z,255);
	n&= 1;
	if (n)
		freopen("stderr.txt","w",stderr);
	fprintf(n ? stderr:stdout,n ? "### %s\n":"## %s\n",
		_dbgTemp+1);
	if (n)
		fflush(stderr);
	//__debugbreak();
	if (n)
		exit(EXIT_FAILURE);
}

void init_window( int width,
	int height )
{
	Rect		bounds;

	remove("stderr");

	rect.right = width;
	rect.bottom = height;
	bounds = rect;
	OffsetRect(&bounds,20,40);

	if ((screen_desktop_window = NewCWindow(NULL,&bounds,"\pNTSC Filter Demo",1,0,
			(WindowRef)-1,0,0))==NULL
			|| NewGWorld(&screen_desktop_port,NES_NTSC_OUT_DEPTH,&rect,
			NULL,NULL,0)<0)
		fatal_error( "video initialization failed" );
}

int read_input( void )
{
	EventRecord	e;
	/* limit to 60 calls per second */
	UnsignedWide	start, now;

	Microseconds(&start);
	//if ( start < next_time && next_time - start > 10000 )
	//	Sleep( (next_time - start)/1000 );
	do
		Microseconds(&now);
	while ( !int64ge(now , next_time) );
	next_time = start;
	int64addu32(*(int64*)&next_time,(int)(1000000/60.0991502455)); /* average NTSC even/odd timing */

	mouse_moved = 0;
	key_pressed = 0;

	while ( GetNextEvent( mDownMask|keyDownMask/*|1<<closeEvent*/,&e ) )
	{
		if ( e.what == mouseDown /*|| e.what == closeEvent*/ )
			return 0;
		if ( e.what == keyDown )
		{
			if ( ((uint16)e.message>>8) == 53 /*MKV_ESCAPE*/ || (uint8)e.message == 'q' )
				return 0;
			key_pressed = (uint8)e.message;
		}
	}
	if ( *(const int32*)&e.where != last_mouse )
	{
		mouse_moved = 1;
		mouse_x = e.where.h / (float) (rect.right - 1) * 2 - 1;
		mouse_y = (1 - e.where.v / (float) (rect.bottom - 1)) * 2 - 1;
	}
	last_mouse = *(const int32*)&e.where;
	return 1;
}

void lock_pixels( void )
{
	GDHandle	old_device;
	CGrafPtr	old_port;
	PixMapHandle	pm = GetGWorldPixMap(screen_desktop_port);

	LockPixels(pm);
	GetGWorld(&old_port,&old_device);
	SetGWorld(screen_desktop_port,NULL);
	PaintRect(&(**pm).bounds);
	SetGWorld(old_port,old_device);
	output_pitch = (**pm).rowBytes&0x3FFF;
	output_pixels = (unsigned char*) (**pm).baseAddr;
}

#if NES_NTSC_OUT_DEPTH>16
 #define COMB_MASK 0x10101
 #define DARK_MASK 0x1F1F1F
#else
 #define COMB_MASK 0x0821
 #define DARK_MASK 0x18E3
#endif

void double_output_height( void )
{
	PixMapHandle	pm = GetGWorldPixMap(screen_desktop_port);
	int y, w = (**pm).bounds.right;

	for ( y = (**pm).bounds.bottom / 2; --y >= 0; )
	{
		unsigned char const* in = output_pixels + y * output_pitch;
		unsigned char* out = output_pixels + y * 2 * output_pitch;
		int n;
		for ( n = w; n; --n )
		{
			unsigned prev = *(nes_ntsc_out_t*) in;
			unsigned next = *(nes_ntsc_out_t*) (in + output_pitch);
			/* mix rgb without losing low bits */
			unsigned mixed = prev + next + ((prev ^ next) & COMB_MASK);
			/* darken by 12% */
			*(nes_ntsc_out_t*) out = prev;
			*(nes_ntsc_out_t*) (out + output_pitch) = (mixed >> 1) - (mixed >> 4 & DARK_MASK);
			in += sizeof(nes_ntsc_out_t);
			out += sizeof(nes_ntsc_out_t);
		}
	}
}

void display_output( void )
{
	PixMapHandle	pm = GetGWorldPixMap(screen_desktop_port);

	CopyBits((BitMap*)*pm,(BitMap*)*GetGWorldPixMap(GetWindowPort(screen_desktop_window)),
		&rect,&rect,ditherCopy,NULL);
	UnlockPixels(pm);
}

void load_image( image_t* out, const uint8 *path, const unsigned char palette [256][3] )
{
	GDHandle	old_device;
	CGrafPtr	old_port;
	CTabHandle	pal;
	ColorSpec	*color;
	CGrafPtr	picp;
	PixMapHandle	pic;
	const uint8	*entry;
	Rect		bounds;
	OpenCPicParams	opp;
	short		refnum;

	ICMProgressProcRecord	progr;

	int		i;

	if (HOpenDF(0,0,path,fsRdPerm,&refnum) || GetPictureFileHeader(refnum,&bounds,&opp)<0)
		goto cfail;
	OffsetRect(&bounds,-bounds.left,-bounds.top);
	if ((pal=(CTabHandle)NewHandleClear(8+sizeof(ColorSpec)*256))==NULL)
		goto cfail;
//	HLock((Handle)pal);
	(**pal).ctSize = 255;
	(**pal).ctFlags = 0x8000;

	i = 0, color = (**pal).ctTable, entry = *palette;
	do
	{
		color->value = i|(i>63 ? 0x4000:0); /* "blacks" beyond this point cannot be used! */
		color->rgb.red = entry[0]*257;
		color->rgb.green = entry[1]*257;
		color->rgb.blue = entry[2]*257;
	} while(++color,entry+=sizeof(*palette),++i<256);
	(**pal).ctSeed = GetCTSeed();

	if (NewGWorld(&picp,8,&bounds,pal,NULL,0)<0)
		goto cfail;

	DisposeCTable(pal);
	pic = GetGWorldPixMap(picp);
	LockPixels(pic);
	GetGWorld(&old_port,&old_device);
	SetGWorld(picp,NULL);
	PaintRect(&(**pic).bounds);

	SetFPos(refnum,fsFromStart,0);
	memset(&progr,0,sizeof(progr));
	if (DrawPictureFile(refnum,&bounds,&progr)<0)
	cfail:	fatal_error( "Couldn't load image" );
	SetGWorld(old_port,old_device);

	FSClose(refnum);

	out->byte_pixels = (unsigned char *) (**pic).baseAddr;
	out->rgb_out      = (nes_ntsc_out_t*) (**pic).baseAddr;
	out->width       = bounds.right;
	out->height      = bounds.bottom;
	out->row_width   = (**pic).rowBytes&0x3FFF;
}

/* * * * * */

/* only used to convert input image to native palette format */
static const unsigned char palette [256][3] = {
	{102,102,102},{  0, 42,136},{ 20, 18,168},{ 59,  0,164},
	{ 92,  0,126},{110,  0, 64},{108,  7,  0},{ 87, 29,  0},
	{ 52, 53,  0},{ 12, 73,  0},{  0, 82,  0},{  0, 79,  8},
	{  0, 64, 78},{  0,  0,  0},{  0,  0,  0},{  0,  0,  0},
	{174,174,174},{ 21, 95,218},{ 66, 64,254},{118, 39,255},
	{161, 27,205},{184, 30,124},{181, 50, 32},{153, 79,  0},
	{108,110,  0},{ 56,135,  0},{ 13,148,  0},{  0,144, 50},
	{  0,124,142},{  0,  0,  0},{  0,  0,  0},{  0,  0,  0},
	{254,254,254},{100,176,254},{147,144,254},{199,119,254},
	{243,106,254},{254,110,205},{254,130,112},{235,159, 35},
	{189,191,  0},{137,217,  0},{ 93,229, 48},{ 69,225,130},
	{ 72,206,223},{ 79, 79, 79},{  0,  0,  0},{  0,  0,  0},
	{254,254,254},{193,224,254},{212,211,254},{233,200,254},
	{251,195,254},{254,197,235},{254,205,198},{247,217,166},
	{229,230,149},{208,240,151},{190,245,171},{180,243,205},
	{181,236,243},{184,184,184},{  0,  0,  0},{  0,  0,  0}
};

#if TARGET_OS_MAC
 #if !TARGET_API_MAC_CARBON
  QDGlobals qd;
 #endif
#endif

int main(int argc,
	char **argv)
{
	image_t image;
	int sony_decoder = 0;
	int merge_fields = 1;
	int burst_phase = 0;
	nes_ntsc_setup_t setup = nes_ntsc_composite;
	nes_ntsc_t* ntsc = (nes_ntsc_t*) malloc( sizeof (nes_ntsc_t) );
	if ( !ntsc )
		fatal_error( "Out of memory" );
(void)argc; (void)argv;
#if TARGET_OS_MAC
	MaxApplZone();
	InitGraf((Ptr)&qd.thePort);
	InitWindows();
	InitMenus();
	TEInit();
	InitDialogs(0);
	{
		CursHandle	cursor;

		if ((cursor=GetCursor(4))==NULL)
			return -1;
		SetCursor(*cursor);
	}
#endif
	{
		FILE	*f;
		void	*p;

		if ((f=fopen("in.pal","rb"))!=NULL)
		{
			if ((p=calloc(512*3,1))!=NULL)
			{
				fread(p,3*512,1,f);
				setup.palette = p;
			}
			fclose(f);
		}
	}
	nes_ntsc_init( ntsc, &setup );
#if TARGET_OS_MAC
	InitCursor();
#endif
	load_image( &image, (/*argc > 1 ? argv [1] :*/ "\ptest.pct"),
		setup.palette ? (const unsigned char(*)[3])setup.palette:palette );

	init_window( NES_NTSC_OUT_WIDTH( image.width ), image.height * 2 );
#if TARGET_OS_MAC
	HideCursor();
#endif
	while ( read_input() )
	{
		lock_pixels();

		burst_phase = (burst_phase+1)%3; //^= 1;
		if ( setup.merge_fields )
			burst_phase = 0;
		
		nes_ntsc_blit( ntsc, image.byte_pixels, image.row_width, burst_phase,
				image.width, image.height, output_pixels, output_pitch );
		
		double_output_height();
		display_output();
		
		switch ( key_pressed )
		{
			case ' ': merge_fields = !merge_fields; break;
			case 'c': setup = nes_ntsc_composite; break;
			case 's': setup = nes_ntsc_svideo; break;
			case 'r': setup = nes_ntsc_rgb; break;
			case 'm': setup = nes_ntsc_monochrome; break;
			case 'd': sony_decoder = !sony_decoder; break;
			default: key_pressed = 0;
		}
		
		if ( key_pressed || mouse_moved )
		{
			setup.merge_fields = merge_fields;
			
			/* available parameters: hue, saturation, contrast, brightness,
			sharpness, gamma, bleed, resolution, artifacts, fringing */
			setup.sharpness = mouse_x;
			setup.gamma     = mouse_y;
			
			setup.decoder_matrix = 0;
			if ( sony_decoder )
			{
				/* Sony CXA2025AS US */
				static float matrix [6] = { 1.630, 0.317, -0.378, -0.466, -1.089, 1.677 };
				setup.decoder_matrix = matrix;
			}
			
			nes_ntsc_init( ntsc, &setup );
		}
	}
	
//	save_bmp( argc > 2 ? argv [2] : "filtered.bmp" );
InitCursor();
	free( ntsc );

	/* write standard 192-byte NES palette */
#if 0
	{
		FILE* out = fopen( "nes.pal", "wb" );
		if ( out )
		{
			unsigned char palette [nes_ntsc_palette_size * 3];
			setup.palette_out = palette;
			nes_ntsc_init( 0, &setup );
			fwrite( palette, 192, 1, out );
		}
	}
#endif
	return 0;
}
