/*
 sound.c
  N2A0x PSG pseuedo-Audio Processing Unit and general sound handlers

 * MacFCEU 0.9a2
 *
 * FCE Ultra - NES/Famicom Emulator
 *
 * Copyright notice for this file:
 *  sound.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>
#include <stdlib.h>

#include "driver.h"
#include "x65xx.h"
#include "FCEU.h"
#include "WAVE.h"
#include "fsound.h"
#include "filter.h"
#include "state.h"

#include "ftrig.h"
#include "memory.h"

#ifdef env68k
 #pragma segment N2A0x
#endif

/* ---- constants */

/* rectangle, triangle */
#define WaveLnMask		0x7FF
#define WaveLnHigh		0x700
/* MESS APU accounts for an initial delay with the triangle channel;
 however, this delay appears to be due to the top flat step with the down-counter.
*/
#define TriLatency		1 //100 /* number of cycles before triangle clocking begins */
/* this is actually a CPU reset delay */
#define FMInitLatency		9

static const uint8	lengthtable[32] =
{ /* updated to fit apu_ref.txt - 0x0D*2 (0x1A) on second row was 0x0e*2...was smoother */
	0x0A, 0xFE, 0x14, 0x02, 0x28, 0x04, 0x50, 0x06,
	0xA0, 0x08, 0x3C, 0x0A, 0x0E, 0x0C, 0x1A, 0x0E,
	0x0C, 0x10, 0x18, 0x12, 0x30, 0x14, 0x60, 0x16,
	0xC0, 0x18, 0x48, 0x1A, 0x10, 0x1C, 0x20, 0x1E
};

/* Dendy UMC UA6527P has #1 and #2 swapped */
int8			RDutyMasks[32];
static uint32		RDutyBitmap = 0x1030FFC;

/* NTSC-based */
static const uint16	GenNoiseFreqTable[16] =
{
	  4,   8,  16,  32,  64,  96, 128, 160,
	202, 254, 380, 508, 762,1016,2034,4068
};
static const uint16	GenDMCTable[16] =
{
	428, 380, 340, 320, 286, 254, 226, 214,
	190, 160, 142, 128, 106,  84,  72,  54
};

static const uint16	PALNoiseFreqTable[16] =
{ /* hard to believe the 7 here is correct */
	  4,   7,  14,  30,  60,  88, 118, 148,
	188, 236, 354, 472, 708, 944,1890,3778
};
static const uint16	PALDMCTable[16] =
{ /* mostly 1/1.08 that of the NTSC table */
	398, 354, 316, 298, 266, 236, 210, 198,
	176, 148, 131, 118,  98,  78,  66,  50,
};

static const uint16	fIRQCounts[2] = {29831, 33255};

/* CPU, mode, count */
static const uint16	fDividerCounts[2][8] =
{
	{ /* 14915*2 4-step total */
		7458, 7458, 7456, 7458, /* merging IRQ steps in two */
		7458+7452, 7458, 7456, 7458  /* merging 5-step no-op */
	},
	{ /* 16627*2 4-step total (was 16626*2 in FCE Ultra) */
		8314, 8314, 8314, 8312, /* merging IRQ steps in one */
		8314+8312, 8314, 8314, 8312 /* merging 5-step no-op */
	}
};
static const uint16	*fDivider = *fDividerCounts;

/* ---- globals */

extern uint32 FEmuSpeedRatio;

/* amplitude tables */
static int32	rectwlookup[32], trilookup[32];
/* 15*9+15*9+127 = 397 is adequate */
static int32	tpcmwlookup[398]; /* triangle/noise/DMC -- increased for overflow, was 203, 400 */

/* virtual wave, prior to filtering */
int32		FCEUWave[13472]; /* 111860.7955 Hz/48/17.3% */
/* post-filter results */
static int16	WaveFinal[13472];

EXPSOUND	GameExpSound; /* zeroed out in FCEU.c */

int32		SndTriActive;
static uint8	TriCount, TriReload;
static int	TriFreq, TriStep;

uint8		PSGReg[16];

uint8		IRQFrameMode;		/* $4017 xx000000 */
uint8		RawDALatch;		/* $4011 0xxxxxxx */

uint8		EnabledChannels;	/* 4015h */
uint8		REnvLoopEnable = REnvelopeLoop; /* set these two to zero for N2A03 rev. A */
uint8		NShShortEnable = NShiftShort;
uint8		DMADualRead; /* older models */
uint8		RNoDutyReset, DADisable, APUSilenceWatch, APUFlatWave;

typedef struct ENVUNIT/*(4)*/
{
	uint8	Speed;
	uint8	reloaddec;//Mode;
	uint8	DecCountTo1;
	uint8	decvolume;
} ENVUNIT;

ENVUNIT		SndEnvUnits[3];

uint8		SndLenCount[4];
const int8	*SndRecDutyTable[2];
int		RecDutyCycle[2];
uint8		RecSweepMask[2];
uint32		RecCurFreq[2];
uint8		RecSweepPeriod[2];

static uint16	nreg;
static int32	sqacc[2];
static int32	tracc;
static int32	rnacc;

uint8		Soundfcnt;
int32		Soundfccnt; /* frame cycle count (decremented) */

uint32		soundtsoffs;

/* multi-rate variables */
int32		nesincsize = 131072;
uint32		soundtsinc = 65536;

/* one of the two tables above loaded by SetSoundVariables */
static uint32 CurrentNoiseFreqTable[16];
uint32		CurrentDMCTable[16];
static uint16	rnvol;

// $4010        -        Frequency
// $4011        -        Actual data outputted
// $4012        -        Address register: $c000 + V*64
// $4013        -        Size register:  Size in bytes = (V+1)*64

int32		APU_DMCacc;
int32		DMCPeriod;
uint8		DMCBitCount;

uint8		DMCAddressLatch, DMCSizeLatch; /* writes to 4012 and 4013 */
uint8		DMCFormat;	/* Write to $4010 */

uint32		DMCAddress;
int32		DMCCounter; /* was DMCSize */

uint8		DMCShiftF[2]; /* first byte kept zero for BE 68K Asm */
#define DMCOutBuf DMCShiftF[1]

uint8		SIRQStat;

uint8		DMCDMABuf;
uint8		DMCBuffered;
uint8		DMCActive;
static uint8	DMCMask;

static void Dummyfunc(void) {};
#if HAVE_HQ
void (*DoRec1)(void)=Dummyfunc;
void (*DoRec2)(void)=Dummyfunc;
void (*DoTriangle)(void)=Dummyfunc;
void (*DoNoise)(void)=Dummyfunc;
void (*DoDMC)(void)=Dummyfunc;
#else
void (*DoRec)(void)=Dummyfunc;
void (*DoPCM)(void)=Dummyfunc;
/* Dendy channels may be entirely separate */
#define DoRec1			DoRec
#define DoRec2			DoRec
#define DoTriangle		DoPCM
#define DoNoise			DoPCM
#define DoDMC			DoPCM
#endif

#if HAVE_HQ
uint32	FCEUChannelBC[5];
#else
uint32	FCEUChannelBC[3];
#endif

static uint16	*ntable;

/* ---- macros */

#define EORNoise(n,c)		(n=((n>>c^n>>14)&1)+n+n)

#define LoadDMCPeriod(V)	(DMCPeriod=CurrentDMCTable[(V)&DMCFMask])

#define PrepDMC() \
	(DMCAddress=0x300+(uint32)DMCAddressLatch<<6,DMCCounter=DMCSizeLatch*16u+1)

/* sweep calculates this every instance */
/* even if sweeping isn't enabled, potential overflow in add mode silences the channel */
#define CheckRFreq(c,s)		(c>=RWaveLnMin && (s&RSweepDir || !(((c>>(s&RSweepMask))+c)&0x800)))

int32		XFStatL;

#if HAVE_HQ
 #define DoRecx(x) \
	if (x) \
		DoRec2(); \
	else \
		DoRec1();
#else
 #define DoRecx(x) \
	DoRec();
#endif

#define RecReload(x,V,t) do{int oc=SndLenCount[x]; \
	if(EnabledChannels&(x+1)) /*(1<<x)*/ { \
		DoRecx(x); \
		if(!RNoDutyReset) \
			RecDutyCycle[x] = 0; \
		if(!(XFStat&16<<x)) /* written on same cycle as envelop decrement */ \
			SndLenCount[x] = lengthtable[V>>RLnCtrShift]; \
	} \
	RecSweepPeriod[x] = (PSGReg[x*4+1]>>kRSweepPeriod&RSweepPeriodMask)+1; \
	SndEnvUnits[x].reloaddec = 1; \
	t=(uint8)RecCurFreq[x], RecCurFreq[x]=t|=(uint32)V<<8&WaveLnHigh; \
	if(!oc && !CheckRFreq(t,PSGReg[x*4+1])) \
		RecSweepMask[x]=0; \
}while(0)

/* ---- prototypes */

DECLFW(WritePSG0);
DECLFW(WritePSG1);
DECLFW(WritePSG2);
DECLFW(WritePSG3);
DECLFW(WritePSG4);
DECLFW(WritePSG5);
DECLFW(WritePSG6);
DECLFW(WritePSG7);
DECLFW(WritePSG); /* 8,9,D */
DECLFW(WritePSGA);
DECLFW(WritePSGB);
DECLFW(WritePSGC);
DECLFW(WritePSGE);
DECLFW(WritePSGF);
DECLFW(WriteDMCCtl);
DECLFW(WriteDMCDA);
DECLFW(WriteDMCAddr);
DECLFW(WriteDMCSize);
DECLFW(APUStatusWrite);
DECLFR(APUStatusRead);
DECLFW(WriteIRQFM);

void SndUpdateRecAB(void);
void SndUpdateTriangle(void);
void SndUpdateRNoise(void);

/* ---- code */

void SndUpdateRecAB(void)
{
	RecSweepMask[0] =-(RecSweepMask[0]&1);
	RecSweepMask[1] =-(RecSweepMask[1]&1);
	SndRecDutyTable[0] = RDutyMasks+(PSGReg[0]>>RDutyTypeShift-3&~7);
	SndRecDutyTable[1] = RDutyMasks+(PSGReg[4]>>RDutyTypeShift-3&~7);
}

void SndUpdateTriangle(void)
{
	SndTriActive =-(FSettings.soundactivech&4 && TriCount && SndLenCount[2]);// && TriFreq>2); //4;
	TriFreq = (PSGReg[10]|((uint32)PSGReg[11]<<8&WaveLnHigh))+1;
}

void SndUpdateRNoise(void)
{
	uint32		v;

	//rnacc&=~1, rnacc|= (int64lo(timestampbase)+timestamp)&1;
	if (!SndLenCount[3] || !(FSettings.soundactivech&8))
		v = 0;
	else if (PSGReg[12]&REnvSpeedAmp) //(SndEnvUnits[2].Mode&1)
		v = SndEnvUnits[2].Speed;
	else
		v = SndEnvUnits[2].decvolume;
	rnvol = v;
}

/* rect A */
DECLFW(WritePSG0)
{
(void)A;
	DoRec1();
	PSGReg[0] = V;
	SndEnvUnits[0].Speed = V&REnvSpeedMask;
	SndRecDutyTable[0] = RDutyMasks+(V>>RDutyTypeShift-3&~7);
}

DECLFW(WritePSG1)
{
(void)A;
	PSGReg[1] = V;
}

DECLFW(WritePSG2)
{
	DoRec1();
	PSGReg[2] = V;
	A = RecCurFreq[0]&WaveLnHigh;
	RecCurFreq[0] = A|=V;
	//if(!CheckRFreq(A,PSGReg[1]))
	//	RecSweepMask[0] = 0;
}

DECLFW(WritePSG3)
{
	RecReload(0,V,A);
	PSGReg[3] = V;
}

/* rect B */
DECLFW(WritePSG4)
{
(void)A;
	DoRec2();
	PSGReg[4] = V;
//	SndEnvUnits[1].Mode = V>>4&3;
	SndEnvUnits[1].Speed = V&REnvSpeedMask;
	SndRecDutyTable[1] = RDutyMasks+(V>>RDutyTypeShift-3&~7);
}

DECLFW(WritePSG5)
{
(void)A;
	PSGReg[5] = V;
	//RecSweepOn[1] = 0;
}

DECLFW(WritePSG6)
{
	DoRec2();
	PSGReg[6] = V;
	A = RecCurFreq[1]&WaveLnHigh;
	RecCurFreq[1] = A|=V;
	//if(!CheckRFreq(A,PSGReg[5]))
	//	RecSweepMask[1] = 0;
}

DECLFW(WritePSG7)
{
	RecReload(1,V,A);
	PSGReg[7] = V;
}

/* 8,9,D */
DECLFW(WritePSG)
{
	PSGReg[A&15] = V;
}

/* triangle */
DECLFW(WritePSGA)
{
(void)A;
	DoTriangle();
	PSGReg[10] = V;
	SndUpdateTriangle();
}

DECLFW(WritePSGB)
{
(void)A;
	DoTriangle();
	if (EnabledChannels&4)
		SndLenCount[2] = lengthtable[V>>3];
	TriReload = 1;
	PSGReg[11] = V;
	SndUpdateTriangle();
}

/* noise */

/* noise mixing test indicates latency */
DECLFW(WritePSGC)
{
(void)A;
	DoNoise();
	PSGReg[12] = V;
	SndEnvUnits[2].Speed = V&REnvSpeedMask;
	SndUpdateRNoise();
}

DECLFW(WritePSGE)
{
(void)A;
	DoNoise();
	PSGReg[14] = V;
}

DECLFW(WritePSGF)
{
(void)A;
	DoNoise();
	if (EnabledChannels&8)
		SndLenCount[3] = lengthtable[V>>3];
	SndEnvUnits[2].reloaddec = 1;
	PSGReg[15] = V;
	SndUpdateRNoise();
}

DECLFW(WriteDMCCtl)
{
(void)A;
	DoDMC();
	LoadDMCPeriod(V);
//	--DMCPeriod; -- breaks test ROMs
	DMCFormat = V;
	if (SIRQStat&DMCSIRQ)
		if (!(V&DMCFEnable))
		{
			XFrame.intlow&=~FCEU_IQDMC; //X6502_IRQEnd
			SIRQStat&=~DMCSIRQ;
		} /*else -- this wasn't found
			XFrame.intlow|= FCEU_IQDMC; //X6502_IRQBegin
		*/
}

DECLFW(WriteDMCDA)
{
(void)A;
	if (DADisable)
		return;
	DoDMC();
	RawDALatch = V&0x7F;
}

DECLFW(WriteDMCAddr)
{
(void)A;
	DMCAddressLatch = V;
}

DECLFW(WriteDMCSize)
{
(void)A;
	DMCSizeLatch = V;
}

void APU_DMCDMA(void)
{ /* DMA reads can occur in the middle of an instruction */
  /* pre-1991 revisions have a bug that causes 401x port reads to be re-read */
	if (DMADualRead && !APU_DMCacc)
		XFDMA|= 0x80;
	DMCDMABuf = X65xx_DMR(DMCAddress,3);
	DMCAddress = (uint16)(DMCAddress+1)|0x8000;
	DMCBuffered = 1;

	if (!--DMCCounter)
		if (DMCFormat&DMCFLoop)
			PrepDMC();
		else if (DMCFormat&DMCFEnable) /* SIRQStat was prev. set anyway */
		{
			SIRQStat|= DMCSIRQ;
			XFrame.intlow|= FCEU_IQDMC; /*X6502_IRQBegin*/
		}
}

DECLFW(APUStatusWrite) //4015h
{ /* this should now be 100% accurate */
(void)A;
#if HAVE_HQ
	DoRec1();
	DoRec2();
	DoTriangle();
	DoNoise();
	DoDMC();
#else
	DoRec();
	DoPCM();
#endif
	SIRQStat&=~DMCSIRQ;
	XFrame.intlow&=~FCEU_IQDMC; /*X6502_IRQEnd*/
	EnabledChannels = V;//&=31;

	/* Force length counters to 0. */
	if (!(V&1))
		SndLenCount[0] = 0;
	if (!(V&2))
		SndLenCount[1] = 0;
	if (!(V&4))
		{SndLenCount[2] = 0, SndTriActive = 0;}
	if (!(V&8))
		SndLenCount[3] = 0;
	if (!(V&16))
		DMCCounter = 0;
	else if (!DMCCounter)
	{
		PrepDMC();
		if (!DMCBuffered)
		{
			XFDMA = 1; /* write cycle for 4015 */
			APU_DMCDMA();
		}
	}
}

uint8 SndGetLenStat(void);
uint8 SndGetLenStat(void)
{
	uint8		ret;

	ret = SndLenCount[3]!=0, ret+= ret;
	ret+= SndLenCount[2]!=0, ret+= ret;
	ret+= SndLenCount[1]!=0, ret+= ret;
	ret+= SndLenCount[0]!=0;
	return ret;
}

DECLFR(APUStatusRead) //4015h
{ /* other than top-precision timing, this is accurate */
	uint8		ret;
(void)A;
	if ((ret=XFStat)&0x40)
		ret&= 15; /* 2A0x emu reflects completion; ret will reflect status on previous cycle */
	else
		ret = SndGetLenStat();
	ret+= (DMCCounter!=0)*16u;
	ret+= SIRQStat;
	if (!fceuindbg)
		SIRQStat&=~FrameSIRQ,
			XFrame.intlow&=~FCEU_IQFCOUNT; /*X6502_IRQEnd*/
	return ret;
}

DECLFW(WriteIRQFM) //4017h
{
(void)A;
	IRQFrameMode = V&FModeMask;
	if (V&FMode5Step)
		{Soundfcnt = 0; APU_FrameUpdate();}
	XFStatL = 0x80000000; /* disable frame IRQ */
	fDivider = fDividerCounts[PAL&1]+((V&FMode5Step)>>7-2);
	Soundfcnt = 1;
	Soundfccnt = fDivider[1]+((int64lo(timestampbase)+timestamp)&1); /* round odd cycle */
	/* no IRQ frame repeat */
	if (!IRQFrameMode) /* neither 5-step nor IRQ disable */
		XFStatL = fIRQCounts[PAL]; /* set IRQ clock */
	else if (V&FModeNoIRQ) /* added no-irq check; 2010.06.02 */
		SIRQStat&=~FrameSIRQ,
			XFrame.intlow&=~FCEU_IQFCOUNT; /*X6502_IRQEnd*/
}

#define sweepIx(P) do{ uint8 swreg; \
	if(!(PSGReg[P*4]&REnvelopeLoop) && SndLenCount[P]) { \
		--SndLenCount[P]; \
		if(XFStat&0x40) \
			XFStat|=16<<P; \
	} \
	/* Frequency Sweep Code Here */ \
	/* xxxx 0000 */ \
	/* xxxx = hz.  120/(P+1) */ \
	swreg=PSGReg[P*4+1]; \
	if(swreg&RSweepEnable) { /* does NOT clock regardless of enabled (test: Donkey Kong) */ \
	 if(!RecSweepPeriod[P] || !--RecSweepPeriod[P]) { int mod; \
		RecSweepPeriod[P]=(swreg>>kRSweepPeriod&RSweepPeriodMask)+1; \
		if(RecCurFreq[P]>=RWaveLnMin /* only one freq check */ \
				&& (mod=swreg&RSweepMask)!=0) { \
			/* putting adder overflow check here produces incorrect results */ \
			mod=RecCurFreq[P]>>mod; \
			if(swreg&RSweepDir) \
				RecCurFreq[P]+=P+~mod; \
			else if (RecSweepMask[P]) /* no check produces distortion in SMB pipe sequence */ \
				RecCurFreq[P]+=mod; \
		} \
	 } \
	} /* APU bug: silence check only at end of sweep operation (including reload) */ \
	RecSweepMask[P]=-CheckRFreq(RecCurFreq[P],swreg); \
}while(0)

//if(SndEnvUnits[P].DecCountTo1>0) \
//	--SndEnvUnits[P].DecCountTo1; \
//if(SndEnvUnits[P].DecCountTo1) \
//	break; \
//SndEnvUnits[P].DecCountTo1=SndEnvUnits[P].Speed+1; \
//if(SndEnvUnits[P].decvolume || SndEnvUnits[P].Mode&2) \

#define envelopeDecayIx(P,R,E) do{ \
	if(SndEnvUnits[P].reloaddec) { \
		SndEnvUnits[P].reloaddec=0,SndEnvUnits[P].decvolume=15; \
		SndEnvUnits[P].DecCountTo1=SndEnvUnits[P].Speed+1; \
		break; \
	} \
	if(SndEnvUnits[P].DecCountTo1 && --SndEnvUnits[P].DecCountTo1) \
		break; \
	SndEnvUnits[P].DecCountTo1=SndEnvUnits[P].Speed+1; \
	if((V=SndEnvUnits[P].decvolume)!=0 || PSGReg[R]&E) \
		SndEnvUnits[P].decvolume=(V-1)&15; \
}while(0)

static void FASTAPASS(1) FrameSoundStuff(uint8 V)
{// Linear counter:  Bit 0-6 of $4008
 // Length counter:  Bit 4-7 of $4003, $4007, $400b, $400f
#if HAVE_HQ
 	DoRec1();
	DoRec2();
	DoNoise();
	DoTriangle();
	DoDMC();
#else
	DoRec();
	DoPCM();
#endif
	if (Soundfccnt==-1)
		XFStat|= SndGetLenStat()|0x40;
	if (!(V&1)) /* Envelope decay, linear counter, length counter, freq sweep */
	{
		sweepIx(0); /* rca */
		sweepIx(1); /* rcb */
		if (!(PSGReg[8]&TStartMask) && SndLenCount[2]) /* tri */
			--SndLenCount[2];
		/* apparently, bit 5 is completely ignored in rev. A */
		if (!(PSGReg[12]&/*REnvelopeLoop*/ REnvLoopEnable) && SndLenCount[3]) /* rns */
			--SndLenCount[3];
	}
	/* Now do envelope decay + linear counter. */
	if (TriReload)
		TriCount = PSGReg[8]&TLoadMask;
	else if (TriCount)
		--TriCount;
	if (!(PSGReg[8]&TStartMask))
		TriReload = 0;
	SndTriActive =-(FSettings.soundactivech&4 && TriCount && SndLenCount[2]);// && TriFreq>2);

	envelopeDecayIx(0,0,REnvelopeLoop);
	envelopeDecayIx(1,4,REnvelopeLoop);
	envelopeDecayIx(2,12,REnvLoopEnable);
	SndUpdateRNoise();
}

#undef sweepIx
#undef envelopeDecayIx

void APU_FrameUpdate(void)
{
	if (!(uint8)((Soundfcnt&3)|IRQFrameMode))
	{
		if (XFStatL>0)
			XFrame.intlow|= FCEU_IQFCOUNT;
		SIRQStat|= FrameSIRQ;
		if (Soundfccnt>=0-2u)
		{ /* repeat flag */
			FrameSoundStuff(Soundfcnt);
			Soundfcnt = 4;
			return;
		}
	}
	if (Soundfcnt!=4)
		FrameSoundStuff(Soundfcnt);
	++Soundfcnt, Soundfcnt&= 3; /* 3 becomes zero, longest in 5-step */
	Soundfccnt+= fDivider[Soundfcnt];
}

/* this might not be necessary, after all... but all attempts to remove it have failed */
void APU_FrameIRQ(void)
{
	XFrame.intlow|= FCEU_IQFCOUNT;
	if (XFStatL<600u)
		XFStatL = 0;
	XFStatL+= fIRQCounts[PAL];
	SIRQStat|= FrameSIRQ;
}

void APU_DMCClockDAC(void)
{
	do
	{
		if (DMCActive)
		{ /* DAC */
			uint8	t;
			/* Unbelievably ugly hack */
			soundtsoffs+= APU_DMCacc; /* non-positive; shouldn't interfere with CPU times */
			DoDMC();
			soundtsoffs-= APU_DMCacc;
			t = (DMCOutBuf<<2&4)-2, DMCOutBuf>>= 1;
			if (!((t+=RawDALatch)&0x80))
				RawDALatch = t;
		}
		--DMCBitCount;
		if (!(DMCBitCount&=7))
		{
			//if (DMCHaveDMA) DMCHaveSample = 1; else...
			if ((DMCActive=DMCBuffered)!=0)
			{
				DMCBuffered = 0;
				DMCOutBuf = DMCDMABuf;
				if (DMCCounter)
					APU_DMCDMA();
			}
		}
	} while((APU_DMCacc+=DMCPeriod)<=0);
}

/* FCEU_SoundCPUHook code moved to fsound.h */

#if HAVE_HQ
/* This has the correct phase.  Don't mess with it. */
static INLINE void RDoRec(int x)
{
	int32		V;
	uint32		amp;
	int32		rthresh;
	int32		*D;
	int32		currdc;
	int32		cf;
	int32		rc;

	if (!(SndLenCount[x]&RecSweepMask[x]))
		amp = 0;
	else if (PSGReg[x*4]&REnvSpeedAmp) //(SndEnvUnits[x].Mode&1)
		amp = SndEnvUnits[x].Speed;
	else
		amp = SndEnvUnits[x].decvolume;
	amp<<= 26;

	//rthresh = RectDuties[PSGReg[x*4]>>6];
	rthresh = PSGReg[x*4]>>6;
	rthresh+= !(rthresh+=rthresh);

	D = &FCEUWave[FCEUChannelBC[x]];
	V = SOUNDTS-FCEUChannelBC[x];

	currdc = RecDutyCycle[x];
	cf = (cf+1)*2u;
	rc = sqacc[x];

	if (V>0)
		do
		{
			if (SndRecDutyTable[x][currdc])
				*D+= amp;
			if (!--rc)
				rc+= cf, currdc = (currdc+1)&7;
		} while(++D,--V);

	RecDutyCycle[x] = currdc;
	sqacc[x] = rc;
endit:	FCEUChannelBC[x] = SOUNDTS;
}

static void RDoRec1(void)
{
	RDoRec(0);
}

static void RDoRec2(void)
{
	RDoRec(1);
}
#endif

/* there may be a mode frequency limit */
#define makettable(x,z) do{const int8 *__rdm; \
	if(!(FSettings.soundactivech&(x+1)) || !(RecSweepMask[x]&SndLenCount[x])) \
		z(RecThreshTable[x][0]=0); \
	else if(PSGReg[x*4]&REnvSpeedAmp) /*(SndEnvUnits[x].Mode&1)*/ \
		RecThreshTable[x][0]=SndEnvUnits[x].Speed; \
	else \
		RecThreshTable[x][0]=SndEnvUnits[x].decvolume; \
	/* if(!inie[x]) RecThreshTable[x][0]=0; // Correct? Buzzing in MM2, others otherwise... */ \
	__rdm = SndRecDutyTable[x]; \
	RecThreshTable[x][7]=RecThreshTable[x][0]&__rdm[7]; \
	RecThreshTable[x][6]=RecThreshTable[x][0]&__rdm[6]; \
	RecThreshTable[x][5]=RecThreshTable[x][0]&__rdm[5]; \
	RecThreshTable[x][4]=RecThreshTable[x][0]&__rdm[4]; \
	RecThreshTable[x][3]=RecThreshTable[x][0]&__rdm[3]; \
	RecThreshTable[x][2]=RecThreshTable[x][0]&__rdm[2]; \
	RecThreshTable[x][1]=RecThreshTable[x][0]&__rdm[1]; \
	RecThreshTable[x][0]&=__rdm[0]; \
} while(0)

/* new high-quality function */
static void RDoRec(void)
{
	int32		V, end;
	int32		freq0, freq1;
	int32		RecThreshTable[2][8];
	int32		totalout;

	V = FCEUChannelBC[0];
	if ((end=SOUNDTS)<=V)
		return;
	FCEUChannelBC[0] = end;

	/* unrolled */
	makettable(0,(void));
	makettable(1,(void));
	freq0=RecCurFreq[0]+1, freq0+= freq0;
	freq1=RecCurFreq[1]+1, freq1+= freq1;
	/* correct implementation shows duty cycles always clock regardless of valid & active */
	totalout = rectwlookup[ RecThreshTable[0][RecDutyCycle[0]] + RecThreshTable[1][RecDutyCycle[1]] ];
	do
	{
		FCEUWave[V>>4]+= totalout;

		--sqacc[1];
		if (--sqacc[0]<=0)
		{
			--RecDutyCycle[0];
			RecDutyCycle[0]&= 7;
			sqacc[0] = freq0; /* set, not add, in case LQ was used */
			if (sqacc[1]<=0)
				goto second; /* dodge unnecessary lookup */
			totalout = rectwlookup[ RecThreshTable[0][RecDutyCycle[0]]
				+RecThreshTable[1][RecDutyCycle[1]] ];
			continue;
		}
		if (sqacc[1]<=0)
		{
		second:	--RecDutyCycle[1];
			sqacc[1] = freq1;
			totalout = rectwlookup[ RecThreshTable[0][RecDutyCycle[0]]
				+RecThreshTable[1][RecDutyCycle[1]&=7] ];
		}
	} while(++V<end);
}

static void RDoRecLQ(void)
{
	int32		V, end;
	int32		freq0, freq1;
	int32		inie[2];
	int32		RecThreshTable[2][8];
	int32		totalout;

	V = FCEUChannelBC[0];
	if ((end=((unsigned)SOUNDTS<<16)/soundtsinc)<=V)
		return;
	FCEUChannelBC[0] = end;

	/* unrolled */
	inie[0]=inie[1]=nesincsize;
	makettable(0,inie[0]=);
	makettable(1,inie[1]=);

	if (!(inie[0]|inie[1]))
		return;
	totalout = rectwlookup[ RecThreshTable[0][RecDutyCycle[0]] + RecThreshTable[1][RecDutyCycle[1]] ];
	freq0=(RecCurFreq[0]+1)<<1+17;
	freq1=(RecCurFreq[1]+1)<<1+17;
	do
	{
		FCEUWave[V>>4]+= totalout; //tmpamp;

		sqacc[1]-= inie[1];
		if ((sqacc[0]-=inie[0])<=0)
		{
			do
				--RecDutyCycle[0];
			while((sqacc[0]+=freq0)<=0);
			RecDutyCycle[0]&= 7;

			if (sqacc[1]<=0)
				goto second; /* dodge unnecessary lookup; 2009.12.27 09:43 EST */
			totalout = rectwlookup[ RecThreshTable[0][RecDutyCycle[0]]
				+RecThreshTable[1][RecDutyCycle[1]] ];
			continue;
		}
		if (sqacc[1]<=0)
		{
		second:	do
				--RecDutyCycle[1];
			while((sqacc[1]+=freq1)<=0);
			RecDutyCycle[1]&= 7;

			totalout = rectwlookup[ RecThreshTable[0][RecDutyCycle[0]]
				+RecThreshTable[1][RecDutyCycle[1]] ];
		}
	} while(++V<end);
}

#if HAVE_HQ
static void RDoTriangle(void)
{
	uint32		V;
	int32		tfreq = TriFreq;
	uint32		trout = trilookup[TriStep&31];
	/* frequency check was missing; 2011.02.10 */
	if (!SndTriActive)
	{ /* Counter is halted, but we still need to output. */
		int32	*start = &FCEUWave[FCEUChannelBC[2]];
		int32	count;

		if ((count=SOUNDTS-FCEUChannelBC[2])!=0)
			do
				*start+++= trout;
			while(--count);
		goto end;
	}

	for(V=FCEUChannelBC[2]; V<SOUNDTS; ++V)
	{
		FCEUWave[V]+= trout;
		if (!--tracc)
		{
			tracc = tfreq;
			trout = trilookup[--TriStep&31];
		}
	}
end:	FCEUChannelBC[2] = SOUNDTS;
}

static void RDoNoise(void)
{
	int32		V, nfreq = CurrentNoiseFreqTable[PSGReg[14]&15];
	int32		outo;
	uint32		namptab[2];
	uint16		nr = nreg;

	namptab[1] = 0;
	namptab[0] = rnvol*FSettings.soundnscale;
	outo = namptab[nr>>14];

	if (PSGReg[14]&NShShortEnable)
		for(V=FCEUChannelBC[3]; V<SOUNDTS; ++V)
		{
			FCEUWave[V]+= outo;
			if (!--rnacc)
				rnacc = nfreq, outo = namptab[(nr=ntable[nr])>>14];
		}
	else
		for(V=FCEUChannelBC[3]; V<SOUNDTS; ++V)
		{
			FCEUWave[V]+= outo;
			if (!--rnacc)
				rnacc = nfreq, outo = namptab[(nr=ntable[nr+32768])>>14];
		}
	nreg = nr;
	FCEUChannelBC[3] = SOUNDTS;
}

static void RDoDMC(void)
{
	uint32		V;

	for(V=FCEUChannelBC[4]; V<SOUNDTS; ++V)
		FCEUWave[V]+= (uint32)RawDALatch<<17;
	FCEUChannelBC[4] = SOUNDTS;
}
#endif

static void RDoTriangleNoiseDMC(void)
{
	const uint16	*nt;
	int32		V, end;
	uint32		trout;
	uint32		nr, nfreq, namptab[2];
	uint32		rnout, dmout;
	int32		totalout;

	V = FCEUChannelBC[2];
	if ((end=SOUNDTS)<=V)
		return;
	FCEUChannelBC[2] = end;

	namptab[1] = 0;
	namptab[0] = rnvol*FSettings.soundnscale;

	trout = trilookup[TriStep&31];
	nr = nreg;
	nt = PSGReg[14]&NShShortEnable ? ntable:ntable+32768;
	rnout = namptab[nr>>14];
	nfreq = CurrentNoiseFreqTable[PSGReg[14]&15];
	dmout = RawDALatch&DMCMask;

	totalout = tpcmwlookup[trout+rnout+dmout];
	if (!namptab[0])
	{
		if (SndTriActive)
		{
			do
			{
				FCEUWave[V>>4]+= totalout;
				if (--tracc<=0)
				{
					tracc = TriFreq;
					trout = trilookup[--TriStep&31];
					totalout = tpcmwlookup[trout+dmout];
				}
				if (--rnacc<=0)
					rnacc = nfreq, nr = nt[nr];
			} while(++V<end);
			V = 0;
		} else
			do
			{
				FCEUWave[V>>4]+= totalout;
				if (--rnacc<=0)
					rnacc = nfreq, nr = nt[nr];
			} while(++V<end);
		goto end;
	}
	if (SndTriActive)
	{
		do
		{
			FCEUWave[V>>4]+= totalout;
			--rnacc;
			if (--tracc<=0)
			{
				tracc = TriFreq;
				trout = trilookup[--TriStep&31];
				if (rnacc>0)
					goto samp;
				goto nsamp;
			}
			if (rnacc<=0)
			{
			nsamp:	rnacc = nfreq, nr = nt[nr];
				rnout = namptab[nr>>14];
			samp:	totalout = tpcmwlookup[trout+rnout+dmout];
			}
		} while(++V<end);
		V = 0;
	} else
		do
		{
			FCEUWave[V>>4]+= totalout;
			if (--rnacc<=0)
			{
				rnacc = nfreq, nr = nt[nr];
				rnout = namptab[nr>>14];
				totalout = tpcmwlookup[trout+rnout+dmout];
			}
		} while(++V<end);
end:	nreg = nr;
}

/* noise channel's always the slowest (optimize as much as possible) */
static void RDoTriangleNoiseDMCLQ(void)
{
	int32		V, end;
	uint32		trout, tfreq, tlen;
	uint32		nlen, namptab[2]; /* namptab[1] is zero */
	uint32		rnout, dmout;
	int32		totalout;

	V = FCEUChannelBC[2];
	if ((end=((uint32)SOUNDTS<<16)/soundtsinc)<=V)
		return;
	FCEUChannelBC[2] = end;

	tlen = (nlen=nesincsize)&SndTriActive;
	tfreq = TriFreq<<16+1;

	namptab[1] = 0;
	if ((namptab[0]=rnvol*FSettings.soundnscale)==0)
		nlen = 0; /* Quick hack speedup, set [nlen] to 0 */
	trout = trilookup[TriStep&31];
	rnout = namptab[nreg>>14&1];
	dmout = RawDALatch&DMCMask;
	totalout = tpcmwlookup[trout+rnout+dmout];

	if (tlen && nlen)
	{
		const uint16	*nt = PSGReg[14]&NShShortEnable ? ntable:ntable+32768;
		uint32		nr = nreg;
		uint32		nfreq = CurrentNoiseFreqTable[PSGReg[14]&15]<<16+1;

		do
		{
			FCEUWave[V>>4]+= totalout;
			rnacc-= nlen;
			tracc-= tlen;
			if (tracc<=0)
			{
				do
					--TriStep;
				while((tracc+=tfreq)<=0);
				trout = trilookup[TriStep&31];
				if (rnacc>0)
					goto samp;
				goto nsamp;
			}
			if (rnacc<=0)
			{
			nsamp:
				do
					nr = nt[nr];
				while((rnacc+=nfreq)<=0);
				rnout = namptab[nr>>14];
			samp:	totalout = tpcmwlookup[trout+rnout+dmout];
			}
		} while(++V<end);
		nreg = nr;
	} else if (tlen)
	{
		do
		{
			FCEUWave[V>>4]+= totalout;
			if ((tracc-=tlen)<=0)
			{
				do
					--TriStep;
				while((tracc+=tfreq)<=0);
				trout = trilookup[TriStep&31];
				totalout = tpcmwlookup[trout+rnout+dmout];
			}
		} while(++V<end);
		V = 0;
	} else if (nlen)
	{
		const uint16	*nt = PSGReg[14]&NShShortEnable ? ntable:ntable+32768;
		uint32		nr = nreg, nfreq = CurrentNoiseFreqTable[PSGReg[14]&15]<<16+1;

		do
		{
			FCEUWave[V>>4]+= totalout;
			rnacc-= nlen;
			if (rnacc<=0)
			{
				do
					nr = nt[nr];
				while((rnacc+=nfreq)<=0);
				rnout = namptab[nr>>14];
				totalout = tpcmwlookup[trout+rnout+dmout];
			}
		} while(++V<end);
		nreg = nr;
	} else
		do
			FCEUWave[V>>4]+= totalout;
		while(++V<end);
}

void APU_MapIO(void)
{
	SetWriteHandler(0x4000,0x4000,WritePSG0);
	SetWriteHandler(0x4001,0x4001,WritePSG1);
	SetWriteHandler(0x4002,0x4002,WritePSG2);
	SetWriteHandler(0x4003,0x4003,WritePSG3);
	SetWriteHandler(0x4004,0x4004,WritePSG4);
	SetWriteHandler(0x4005,0x4005,WritePSG5);
	SetWriteHandler(0x4006,0x4006,WritePSG6);
	SetWriteHandler(0x4007,0x4007,WritePSG7);
	SetWriteHandler(0x4008,0x400D,WritePSG);
	SetWriteHandler(0x400A,0x400A,WritePSGA);
	SetWriteHandler(0x400B,0x400B,WritePSGB);
	SetWriteHandler(0x400C,0x400C,WritePSGC);
	SetWriteHandler(0x400E,0x400E,WritePSGE);
	SetWriteHandler(0x400F,0x400F,WritePSGF);
	SetWriteHandler(0x4010,0x4010,WriteDMCCtl);
	SetWriteHandler(0x4011,0x4011,WriteDMCDA);
	SetWriteHandler(0x4012,0x4012,WriteDMCAddr);
	SetWriteHandler(0x4013,0x4013,WriteDMCSize);
	SetReadHandler(0x4015,0x4015,APUStatusRead);
	SetWriteHandler(0x4015,0x4015,APUStatusWrite);
	SetWriteHandler(0x4017,0x4017,WriteIRQFM);
}

static int32	inbuf = 0;

/* returns 1 if the samples are entirely equal to the first of the previous frame */
static int ScanForSilence(void)
{
	static int32	foresample;
	int32		*s, c, f;
	int		r = 0;

	if ((c=inbuf)<=0)
		return 0;
	s = FCEUWave, f = s[c-1];
	if (c&1 && *s++!=foresample)
		goto end;
	if ((c&2 && (*s++-foresample)|(*s++-foresample)) || !(c&=0x7FFFFFFC))
		goto end;
	do
		if ((*s++-foresample)|(*s++-foresample)|(*s++-foresample)|(*s++-foresample))
			goto end;
	while(c-=4);
	r = 1;
end:	foresample = f;
	return r;
}

int FlushEmulateSound(void)
{
	uint32		b;
	int32		x, end, left;

	if ((x=timestamp)==0)
		return 0;
	if (!FSettings.SndRate)
	{
		inbuf = left = end = 0;
		goto nosnd;
	}
#if HAVE_HQ
	DoRec1();
	DoRec2();
	DoTriangle();
	DoNoise();
	DoDMC();
	if (FSettings.soundq)
	{
		int32	*s = &FCEUWave[soundtsoffs];

		if (FSettings.soundactivech&0x20)
			for(b=0; b<=SCSS5B; ++b)
				if (FSettings.soundactivech&(64u<<b) && GameExpSound.HiFill[b])
					GameExpSound.HiFill[b]();
		/* ==0 already checked */
		do
			b = *s, *s++ = (b&0x1FFFF)+tpcmwlookup[b>>17&0x1FF]+rectwlookup[b>>26];
		while(--x);
		end = NeoFilterSound(FCEUWave,WaveFinal,SOUNDTS,&left);
		inbuf = end;

		memmove(FCEUWave,FCEUWave+SOUNDTS-left,left*sizeof(uint32));
		memset(FCEUWave+left,0,sizeof(FCEUWave)-left*sizeof(uint32));

		for(b=0; b<=SCSS5B; ++b)
			if (GameExpSound.HiSync[b])
				GameExpSound.HiSync[b](left);

		FCEUChannelBC[0] = FCEUChannelBC[1] = FCEUChannelBC[2] =
			FCEUChannelBC[3] = FCEUChannelBC[4] = left;
	}
	else
#else
	DoRec();
	DoPCM();
#endif
	{
		end = ((uint32)SOUNDTS<<16)/soundtsinc;
		if (FSettings.soundactivech&0x20)
			for(b=0; b<=SCSS5B; ++b)
				if (FSettings.soundactivech&(64u<<b) && GameExpSound.Fill[b])
					GameExpSound.Fill[b](end&15);
		inbuf = end>>4;
		if (APUSilenceWatch)
			APUFlatWave = ScanForSilence();
		SexyFilter(FCEUWave,WaveFinal,inbuf);

		//if(FSettings.lowpass)
		// SexyFilter2(WaveFinal,end>>4);

		if (end&15) /* roll ending sample to beginning */
			FCEUWave[0] = FCEUWave[inbuf];
		FCEUWave[inbuf] = 0;
	}
nosnd:
#if HAVE_HQ
	if (FSettings.soundq)
		soundtsoffs = left;
	else
	{
		FCEUChannelBC[4] = FCEUChannelBC[3] = FCEUChannelBC[1] =
#endif
			FCEUChannelBC[2] = FCEUChannelBC[0] = end&15;
		soundtsoffs = (soundtsinc*(end&15))>>16;
#if HAVE_HQ
	}
#endif
	FCEU_WriteWaveData(WaveFinal,inbuf); /* does nothing if sound recording is off */
	return inbuf;
}

int16 *FCEU_GetSoundBuffer(int32 *c)
{
	*c = inbuf;
	return WaveFinal;
}

/* FIXME:  Find out what sound registers get reset on reset.  I know $4001/$4005 don't,
due to that whole MegaMan 2 Game Genie thing.
*/

void APU_Reset(void)
{
	fDivider = fDividerCounts[PAL&1]+((IRQFrameMode&FMode5Step)>>7-2); /* IRQFrameMode unchanged */
	Soundfcnt = 1;
	Soundfccnt = fDivider[1]-(FMInitLatency-1);

	nreg = 1;
	if (!nesincsize)
		nesincsize = 131072;
	rnacc = sqacc[1] = sqacc[0] = (2048u<<17)/nesincsize;
	tracc = TriLatency;

	//RecSweepOn[0] = RecSweepOn[1] = 0;
	RecSweepPeriod[0] = RecSweepPeriod[1] = 0;
	RecCurFreq[0] = RecCurFreq[1] = 0;

	/* some of these are cleared only on Power */
	DMCAddressLatch = DMCSizeLatch = DMCFormat = DMCOutBuf = DMCDMABuf = 0;
	DMCCounter = DMCBuffered = DMCActive = 0;
	DMCAddress = 0xC000;
	APU_DMCacc = CurrentDMCTable[0]; //1; 2011.02.13
	DMCBitCount = 0;

	SIRQStat = 0;
	XFStatL = 0x80000000u;

	RawDALatch = 0;

	TriCount = 0;
	TriReload = 0;
	TriStep = 0;
	EnabledChannels = 0;

	SndLenCount[0] = SndLenCount[1] = SndLenCount[2] = SndLenCount[3] = 0;
	SndTriActive = 0;

	/* added; 2011.02.10 */
	WritePSG3(0x4003,0); /* wrote to the wrong memory with basic WritePSG */
	WritePSG7(0x4007,0);
	PSGReg[0] = PSGReg[4] = PSGReg[12] = 16;
	SndEnvUnits[0].reloaddec = SndEnvUnits[1].reloaddec =
		SndEnvUnits[2].reloaddec = 1;
}

void APU_Power(void)
{
	APU_MapIO();
	dwmemset(PSGReg,0,sizeof(PSGReg)); /* length counters are erased */
	dwmemset(FCEUWave,0,sizeof(FCEUWave));
	dwmemset(SndEnvUnits,0,sizeof(SndEnvUnits));
	APU_Reset();
	//++Soundfccnt; /* reset is slightly longer */
	/* mode is not initialized on reset */
	IRQFrameMode = FEmDendy ? FModeNoIRQ:0; /* enabled for standard case */
	/* AV model has different reset features, apparently, that it fails the test ROMs */
#if HAVE_HQ
	FCEUChannelBC[4] = FCEUChannelBC[3] = FCEUChannelBC[1] =
#endif
		FCEUChannelBC[2] = FCEUChannelBC[0] = 0;

	soundtsoffs = 0;
	LoadDMCPeriod(DMCFormat);
}

static void SndSetFuncs(uint8 bits)
{
#if HAVE_HQ
	if (FSettings.soundq) /* >=0 */
	{
		OptFAddr(DoRec1,bits&1 ? RDoRec1:Dummyfunc);
		OptFAddr(DoRec2,bits&2 ? RDoRec2:Dummyfunc);
		OptFAddr(DoTriangle,bits&4 ? RDoTriangle:Dummyfunc);
		OptFAddr(DoNoise,bits&8 ? RDoNoise:Dummyfunc);
		OptFAddr(DoDMC,bits&16 ? RDoDMC:Dummyfunc);
		return;
	} else
#endif
	{
		uint8		bq = soundtsinc==65536;
#if HAVE_HQ
		OptFAddr(DoRec1,bits&1 ? bq ? RDoRec:RDoRecLQ:Dummyfunc);
		OptFAddr(DoRec2,bits&2 ? bq ? RDoRec:RDoRecLQ:Dummyfunc);
		OptFAddr(DoTriangle,bits&4 ? bq ? RDoTriangleNoiseDMC:RDoTriangleNoiseDMCLQ:Dummyfunc);
		OptFAddr(DoNoise,bits&8 ? bq ? RDoTriangleNoiseDMC:RDoTriangleNoiseDMCLQ:Dummyfunc);
		OptFAddr(DoDMC,bits&16 ? bq ? RDoTriangleNoiseDMC:RDoTriangleNoiseDMCLQ:Dummyfunc);
#else
		OptFAddr(DoRec,bits&3 ? bq ? RDoRec:RDoRecLQ:Dummyfunc);
		OptFAddr(DoPCM,bits&28 ? bq ? RDoTriangleNoiseDMC:RDoTriangleNoiseDMCLQ:Dummyfunc);
#endif
	}
}

static void LoadLUTs(void)
{
	const uint16	*nc, *dc;
	uint32		*na = CurrentNoiseFreqTable, *da = CurrentDMCTable;
	int32		x, t;

	if (PAL)
		nc = PALNoiseFreqTable, dc = PALDMCTable;
	else
		nc = GenNoiseFreqTable, dc = GenDMCTable;
	x = 16;
	do
	{
		*na++ = *nc++;
		if (RateFixFlags&0x40) /* now optional; 2013.05.15 */
			*da++ = *dc++;
		else
			*da++ = umuldiv(*dc++,FEmuSpeedRatio,500000000)+1>>1;
	} while(--x);

	x = 0, t = RDutyBitmap;
	do
		RDutyMasks[x] = t>>31;
	while(t+=t,++x<32);
}

static int BuildWLUTs(void)
{
	double		numer;
	int		x;

	LoadLUTs();
	/* speed boost: noise table */
	if (!ntable && (ntable=malloc(32768*2*sizeof(uint16)))==NULL)
	{ /* no memory for sound, for that matter... */
		FSettings.SndRate = 0;
		return 0;
	}
	x = 0;
	do
	{
		ntable[x] = (((x>>8^x>>14)&1)+x+x)&0x7FFF;
		ntable[x+32768] = (((x>>13^x>>14)&1)+x+x)&0x7FFF;
	} while(++x<32768);

	{ /* triangle scale table */
		int32		v, n = FSettings.soundtscale;
	#if HAVE_HQ
		if (FSettings.soundq)
			n<<= 17;
	#endif
		x = 0, v = n*(sizeof(trilookup)/sizeof(trilookup[0])/2);
		do
			trilookup[x] = v-=n;
		while(++x<sizeof(trilookup)/sizeof(trilookup[0])/2);
		do
			trilookup[x] = v, v+=n;
		while(++x<sizeof(trilookup)/sizeof(trilookup[0]));
	}

	/* rectangle filter */
	x = 1;
	numer = (95.52*64)*FSettings.soundrectamp;
	do
		rectwlookup[x] = numer/(8128.0/x+100);
	while(++x<32);
	/* triangle/noise/DMC filter */
	x = 1, numer = (163.67*64)*FSettings.soundtndamp;
	do
		tpcmwlookup[x] = numer/(24329.0/x+100);
	while(++x<sizeof(tpcmwlookup)/sizeof(tpcmwlookup[0]));

	return 1;
}

static void CalcIncSize(void)
{
	soundtsinc = (unsigned)nesincsize;
	if (FEmuSpeedRatio!=1000000000UL && !(RateFixFlags&0x80))
		soundtsinc = umuldiv(soundtsinc,FEmuSpeedRatio,1000000000UL);
	soundtsinc>>= 1;
}

void FCEU_SetSoundVariables(void)
{
extern uint8 FCEU_soundlogrec;
	uint64		ratio;
	uint32		Rate, maxrate;
	int		x;

	if (!FEmuSpeedRatio)
		FEmuSpeedRatio = 1000000000UL;

	rectwlookup[0] = tpcmwlookup[0] = 0;

	if ((Rate=FSettings.SndRate)!=0)
	{ /* build the wave amplitude LUTs */
		if (!BuildWLUTs())
			goto silence;
	} else
	{
	silence:
		SndSetFuncs(0);
		return;
	}
	if (Rate<32)
		Rate = 32;
#if HAVE_HQ
	MakeFilters(Rate>>4);
#endif
	maxrate = FCEU_GetCPUBaseRate(&ratio);
	if (!FCEU_soundlogrec && Rate>maxrate)
		FSettings.SndRate = Rate = maxrate;
	fDivider = fDividerCounts[PAL&1]+((IRQFrameMode&FMode5Step)>>5);

	for(x=0; x<=SCSS5B; ++x)
		if (GameExpSound.RChange[x])
			GameExpSound.RChange[x]();
	sqacc[0] = sqacc[1] = 0;
#if HAVE_HQ
	FCEUChannelBC[4] = FCEUChannelBC[3] = FCEUChannelBC[1] =
#endif
		FCEUChannelBC[2] = FCEUChannelBC[0] = 0;

	LoadDMCPeriod(DMCFormat); /* in case of changing from PAL to NTSC */
#if HAVE_HQ
	if (FSettings.soundq)
		nesincsize = 131072;
	else
#endif
	nesincsize = uint64divu32(ratio,Rate);
	CalcIncSize();
	SndSetFuncs(FSettings.soundactivech);

	rnacc = sqacc[1] = sqacc[0] = (2048u<<17)/nesincsize;
	tracc = TriLatency;

	/* recalculate original rate (absolutely needed in case of rate clip) */
	FSettings.OrgRate = FSettings.SndRate;
	if (RateFixFlags&0x80)
		FSettings.OrgRate = umuldiv(FSettings.SndRate,FEmuSpeedRatio,1000000000UL);
}

void FCEUI_SetDutyBitmap(uint32 bitmap)
{
	RDutyBitmap = bitmap;
	LoadLUTs();
}

void FCEUI_SetDutyReset(int r)
{
	RNoDutyReset = !r;
}

void FCEUI_SetDirectDADisable(int r)
{
	DADisable = !!r;
}

int FCEUI_Sound(int Rate)
{
	if ((Rate<<4>>4)!=Rate)
		return 0;
	return FCEUI_Sound2(Rate<0 ? 0:Rate<<4);
}

int32 FCEUI_Sound2(int32 Rate)
{
	if (Rate-1u>=(uint32)NTSC_CPU-1u)
		Rate = (uint32)NTSC_CPU&~(Rate>>31);
	FSettings.OrgRate = Rate;
	/* removed 1.6MHz adjustment code */
	/* emulated rate; may be adjusted */
	FSettings.SndRate = Rate;
	FCEU_SetSoundVariables();
	Rate = FSettings.OrgRate;
	FCEUI_SetDCFilterLoss(FSettings.sounddcfl);
	return Rate;
}

/* added; 2008.09.12 */
void FCEUI_SetSoundActiveChannels(uint16 bits)
{ /* flush before setting */
#if HAVE_HQ
	DoRec1();
	DoRec2();
	DoTriangle();
	DoNoise();
#else
	DoRec();
	DoPCM();
#endif
	FSettings.soundactivech = bits;

	if (FSettings.SndRate)
		SndSetFuncs(bits);
	SndUpdateTriangle();
	SndUpdateRNoise();
	DMCMask =-(bits>>4&1);
}

#if HAVE_HQ
void FCEUI_SetLowPass(int q)
{
	FSettings.lowpass = q;
}

void FCEUI_SetSoundQuality(int qty)
{
	FSettings.soundq = (unsigned)qty>255 ? 255:qty;
	FCEU_SetSoundVariables();
}
#endif

void FCEUI_SetSoundVolume(int32 vol)
{
	FSettings.SndVolF = (int32)((uint32)(FSettings.SoundVolume=vol)
		<<(
	#if HAVE_HQ
		FSettings.soundq ? 15:
	#endif
		15-4))/25;
}

/* added; 2010.03.23 */
void FCEUI_SetRectFilterAmp(int16 amp)
{
	FSettings.soundrectamp = amp;
	FCEU_SetSoundVariables();
}

/* added; 2013.05.04 */
void FCEUI_SetTNDFilterAmp(int16 amp)
{
	FSettings.soundtndamp = amp;
	BuildWLUTs();
	CalcIncSize();
}

/* added; 2010.07.17 */
void FCEUI_SetNoiseScale(uint16 scale)
{ /* internal multiplier (no LUT) */
	FSettings.soundnscale = scale;
	SndUpdateRNoise();
}

/* added; 2013.05.05 */
void FCEUI_SetTriangleScale(uint16 scale)
{
	FSettings.soundtscale = scale;
	BuildWLUTs();
}

void FCEUI_SetDCFilterLoss(int loss)
{
	if (FSettings.OrgRate<4)
		FSettings.OrgRate = 4;
	if (loss<0)
		loss = 0;
	else if (loss>127)
		loss = 127;
	FSettings.sounddcfl = loss;
	FSettings.SndRateF = (loss<<16+9)/(FSettings.OrgRate>>2);
}

SFORMAT FCEUAPU_STATEINFO[] =
{
	{&Soundfccnt,4|FCEUSTATE_RLSB,"FHCN"},
	{&Soundfcnt,1,"FCNT"},
	{ PSGReg,16,"PSG"},
	{&EnabledChannels,1,"ENCH"},
	{&IRQFrameMode,1,"IQFM"},
	{&nreg,2|FCEUSTATE_RLSB,"NREG"},
	{&TriReload,1,"TRIM"},
	{&TriCount,1,"TRIC"},

	{&SndEnvUnits[0].Speed,1,"E0SP"},
	{&SndEnvUnits[1].Speed,1,"E1SP"},
	{&SndEnvUnits[2].Speed,1,"E2SP"},

	{&SndEnvUnits[0].reloaddec,1,"E0RL"},
	{&SndEnvUnits[1].reloaddec,1,"E1MO"},
	{&SndEnvUnits[2].reloaddec,1,"E2MO"},

	{&SndEnvUnits[0].DecCountTo1,1,"E0D1"},
	{&SndEnvUnits[1].DecCountTo1,1,"E1D1"},
	{&SndEnvUnits[2].DecCountTo1,1,"E2D1"},

	{&SndEnvUnits[0].decvolume,1,"E0DV"},
	{&SndEnvUnits[1].decvolume,1,"E1DV"},
	{&SndEnvUnits[2].decvolume,1,"E2DV"},

	{&SndLenCount[0],1,"LEN0"},
	{&SndLenCount[1],1,"LEN1"},
	{&SndLenCount[2],1,"LEN2"},
	{&SndLenCount[3],1,"LEN3"},
	{ RecSweepMask,2,"SWEE"},
	{&RecCurFreq[0],4|FCEUSTATE_RLSB,"CRF1"},
	{&RecCurFreq[1],4|FCEUSTATE_RLSB,"CRF2"},
	{ RecSweepPeriod,2,"SWCT"},

	{&SIRQStat,1,"SIRQ"},
	{&XFStatL,4|FCEUSTATE_RLSB,"SFSL"},

	{&APU_DMCacc,4|FCEUSTATE_RLSB,"5ACC"},
	{&DMCBitCount,1,"5BIT"},
	{&DMCAddress,4|FCEUSTATE_RLSB,"5ADD"},
	{&DMCCounter,4|FCEUSTATE_RLSB,"5SIZ"},
	{&DMCShiftF,2,"5SHF"},

	{&DMCBuffered,1,"5HVD"}, //truncated from 5HVDM, 5HVSP
	{&DMCActive,1,"5HVS"},

	{&DMCSizeLatch,1,"5SZL"},
	{&DMCAddressLatch,1,"5ADL"},
	{&DMCFormat,1,"5FMT"},
	{&RawDALatch,1,"RWDA"},

//	{&sqacc[0],1,"SQA0"},
//	{&sqacc[1],1,"SQA1"},
	{0}
};

void FCEUAPU_SaveState(void)
{
}

void FCEUAPU_LoadState(int v)
{
	if (v<9819) /* prev. downshifted 6 */
		IRQFrameMode<<= 6;
	else if (v<9821)
		Soundfcnt>>= 6;
	else if (v<9822)
		DMADualRead = 0, DMCBitCount = ~DMCBitCount&7;
	Soundfcnt&= 3;
	LoadDMCPeriod(DMCFormat);
	DMCAddress = (uint16)DMCAddress|0x8000;
	RawDALatch&= 127;
	DMCShiftF[0] = 0; //for an Asm optimization
	RecCurFreq[0]&= WaveLnMask, RecCurFreq[1]&= WaveLnMask;

	SndUpdateTriangle();
	SndUpdateRecAB();
	SndUpdateRNoise();
}

uint8 FCEUAPU_RegPeek(uint16 A)
{ /* bit ordering is kind of made up outside logical register range */
	int		x = A>>2&3;

	switch(A)
	{
	case 2:
	case 6: return (uint8)RecCurFreq[x];
	case 3:
	case 7: return RecCurFreq[x]>>8|(PSGReg[A]&~7);
	case 0x10: return DMCFormat;
	case 0x11: return (uint8)(RawDALatch|DMCBuffered<<7);
	case 0x12: return DMCAddressLatch;
	case 0x13: return DMCSizeLatch;
	case 0x14: return DMCOutBuf;
	case 0x15: return EnabledChannels;
	case 0x16: return (uint8)DMCCounter;
	case 0x17: return DMCCounter>>8|DMCActive<<7|(DMCBitCount<<4);
	case 0x18: return (uint8)DMCAddress;
	case 0x19: return (uint8)(DMCAddress>>8);
	case 0x1A: return (uint8)APU_DMCacc;
	case 0x1B: return (uint8)(APU_DMCacc>>8);
	case 0x1C:
	case 0x1D:
	case 0x1E:
	case 0x1F: return SndLenCount[A&3];

	case 0x20:
	case 0x24:
	case 0x2B: return SndEnvUnits[x].decvolume<<4|SndEnvUnits[x].DecCountTo1;
	case 0x21:
	case 0x25: return (uint8)(RecSweepMask[x]<<7|(RecDutyCycle[x]<<4&0x70)|RecSweepPeriod[x]);
	case 0x22:
	case 0x26: return (uint8)sqacc[x];
	case 0x23:
	case 0x27: return (uint8)(sqacc[x]>>8);
	case 0x28: return (uint8)(TriCount|TriReload<<7);
	case 0x29: return (uint8)tracc;
	case 0x2A: return (uint8)(tracc>>8|TriStep<<3);
	case 0x2C: return (uint8)nreg;
	case 0x2D: return (uint8)(nreg>>8);
	case 0x2E: return (uint8)rnacc;
	case 0x2F: return (uint8)(rnacc>>8);

	default: return PSGReg[A&15];
	}
}

void FCEUAPU_RegPoke(uint16 A,
	uint8 V,
	int x)
{
	x = A>>2&3;
	switch(A)
	{
	case 2:
	case 6: RecCurFreq[x] = V|(RecCurFreq[x]&WaveLnHigh); goto set;
	case 3:
	case 7: RecCurFreq[x] = V<<8&WaveLnHigh|(uint8)RecCurFreq[x]; goto set;

	case 0x10: DMCFormat = V; break;
	case 0x11: RawDALatch = V&0x7F, DMCBuffered = V>>7; break;
	case 0x12: DMCAddressLatch = V; break;
	case 0x13: DMCSizeLatch = V; break;
	case 0x14: DMCOutBuf = V; break;
	case 0x15: EnabledChannels = V; break;
	case 0x16: DMCCounter = V|(DMCCounter&0xF00); break;
	case 0x17:
		DMCCounter = (V<<8&0xF00)|(uint8)DMCCounter;
		DMCActive = V>>7;
		DMCBitCount = V>>4&7;
		break;
	case 0x18: DMCAddress = V|(DMCAddress&0xFF00); break;
	case 0x19: DMCAddress = V<<8|(uint8)DMCAddress; break;
	case 0x1A: APU_DMCacc = V|(APU_DMCacc&0xFF00); break;
	case 0x1B: APU_DMCacc = V<<8|(uint8)APU_DMCacc; break;
	case 0x1C:
	case 0x1D:
	case 0x1E:
	case 0x1F: SndLenCount[A&3] = V; break;

	case 0x20:
	case 0x24:
	case 0x2B:
		SndEnvUnits[x].decvolume = V>>4;
		SndEnvUnits[x].DecCountTo1 = V&15;
		break;
	case 0x21:
	case 0x25:
		RecSweepMask[x] = V>>7;
		RecDutyCycle[x] = V>>4&7;
		RecSweepPeriod[x] = V&15;
		break;
	case 0x22:
	case 0x26: sqacc[x] = V|(sqacc[x]&0xFF00); break;
	case 0x23:
	case 0x27: sqacc[x] = V<<8|(uint8)sqacc[x]; break;
	case 0x28: TriCount = V&0x7F, TriReload = V>>7; break;
	case 0x29: tracc = V|(tracc&0xFF00); break;
	case 0x2A: tracc = V<<8|(uint8)tracc, TriStep = V>>3; break;
	case 0x2C: nreg = V|(nreg&0xFF00); break;
	case 0x2D: nreg = V<<8|(uint8)nreg; break;
	case 0x2E: rnacc = V|(rnacc&0xFF00); break;
	case 0x2F: rnacc = V<<8|(uint8)rnacc; break;

	default: set: PSGReg[A&15] = V;
	}
	SndUpdateRecAB();
	SndUpdateTriangle();
	SndUpdateRNoise();
}
